# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings

import tensorflow as tf
from tensorflow import keras


class BinaryCrossentropy(keras.losses.Loss):
    """Computes the cross-entropy loss between true labels and predicted labels.

    Use this cross-entropy loss for binary (0 or 1) classification applications.
    This loss is updated for YoloX by offering support for no axis to mean over.

    Args:
        from_logits: Whether to interpret `y_pred` as a tensor of
            [logit](https://en.wikipedia.org/wiki/Logit) values. By default, we
            assume that `y_pred` contains probabilities (i.e., values in [0,
            1]).
        label_smoothing: Float in [0, 1]. When 0, no smoothing occurs. When >
            0, we compute the loss between the predicted labels and a smoothed
            version of the true labels, where the smoothing squeezes the labels
            towards 0.5.  Larger values of `label_smoothing` correspond to
            heavier smoothing.
        axis: the axis along which to mean the ious. Defaults to `no_reduction`
            which implies mean across no axes.

    Example:
    ```python
    model.compile(
      loss=BinaryCrossentropy(from_logits=True)
      ....
    )
    ```
    """

    def __init__(
        self, from_logits=False, label_smoothing=0.0, axis=None, **kwargs
    ):
        super().__init__(**kwargs)
        self.from_logits = from_logits
        self.label_smoothing = label_smoothing
        self.axis = axis

    def call(self, y_true, y_pred):
        y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)
        label_smoothing = tf.convert_to_tensor(
            self.label_smoothing, dtype=y_pred.dtype
        )

        def _smooth_labels():
            return y_true * (1.0 - label_smoothing) + 0.5 * label_smoothing

        y_true = tf.__internal__.smart_cond.smart_cond(
            label_smoothing, _smooth_labels, lambda: y_true
        )

        if self.axis == "no_reduction":
            warnings.warn(
                "`axis='no_reduction'` is a temporary API, and the API"
                "contract will be replaced in the future with a more generic "
                "solution covering all losses."
            )
            return tf.reduce_mean(
                keras.backend.binary_crossentropy(
                    y_true, y_pred, from_logits=self.from_logits
                ),
                axis=self.axis,
            )

        return keras.backend.binary_crossentropy(
            y_true, y_pred, from_logits=self.from_logits
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "from_logits": self.from_logits,
                "label_smoothing": self.label_smoothing,
                "axis": self.axis,
            }
        )
        return config



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_cv.src.core.factor_sampler.constant_factor_sampler import (
    ConstantFactorSampler,
)
from keras_cv.src.core.factor_sampler.factor_sampler import FactorSampler
from keras_cv.src.core.factor_sampler.normal_factor_sampler import (
    NormalFactorSampler,
)
from keras_cv.src.core.factor_sampler.uniform_factor_sampler import (
    UniformFactorSampler,
)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""RegNet models for KerasCV.
References:
    - [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678) (CVPR 2020)
    - [Based on the Original keras.applications RegNet](https://github.com/keras-team/keras/blob/master/keras/applications/regnet.py)
"""  # noqa: E501

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import backend
from tensorflow.keras import layers

from keras_cv.src.layers import SqueezeAndExcite2D
from keras_cv.src.models.legacy import utils
from keras_cv.src.models.legacy.weights import parse_weights

# The widths and depths are deduced from a quantized linear function. For
# more information, please refer to "Designing Network Design Spaces" by
# Radosavovic et al.

# BatchNorm momentum and epsilon values taken from original implementation.

MODEL_CONFIGS = {
    "x002": {
        "depths": [1, 1, 4, 7],
        "widths": [24, 56, 152, 368],
        "group_width": 8,
        "default_size": 224,
        "block_type": "X",
    },
    "x004": {
        "depths": [1, 2, 7, 12],
        "widths": [32, 64, 160, 384],
        "group_width": 16,
        "default_size": 224,
        "block_type": "X",
    },
    "x006": {
        "depths": [1, 3, 5, 7],
        "widths": [48, 96, 240, 528],
        "group_width": 24,
        "default_size": 224,
        "block_type": "X",
    },
    "x008": {
        "depths": [1, 3, 7, 5],
        "widths": [64, 128, 288, 672],
        "group_width": 16,
        "default_size": 224,
        "block_type": "X",
    },
    "x016": {
        "depths": [2, 4, 10, 2],
        "widths": [72, 168, 408, 912],
        "group_width": 24,
        "default_size": 224,
        "block_type": "X",
    },
    "x032": {
        "depths": [2, 6, 15, 2],
        "widths": [96, 192, 432, 1008],
        "group_width": 48,
        "default_size": 224,
        "block_type": "X",
    },
    "x040": {
        "depths": [2, 5, 14, 2],
        "widths": [80, 240, 560, 1360],
        "group_width": 40,
        "default_size": 224,
        "block_type": "X",
    },
    "x064": {
        "depths": [2, 4, 10, 1],
        "widths": [168, 392, 784, 1624],
        "group_width": 56,
        "default_size": 224,
        "block_type": "X",
    },
    "x080": {
        "depths": [2, 5, 15, 1],
        "widths": [80, 240, 720, 1920],
        "group_width": 120,
        "default_size": 224,
        "block_type": "X",
    },
    "x120": {
        "depths": [2, 5, 11, 1],
        "widths": [224, 448, 896, 2240],
        "group_width": 112,
        "default_size": 224,
        "block_type": "X",
    },
    "x160": {
        "depths": [2, 6, 13, 1],
        "widths": [256, 512, 896, 2048],
        "group_width": 128,
        "default_size": 224,
        "block_type": "X",
    },
    "x320": {
        "depths": [2, 7, 13, 1],
        "widths": [336, 672, 1344, 2520],
        "group_width": 168,
        "default_size": 224,
        "block_type": "X",
    },
    "y002": {
        "depths": [1, 1, 4, 7],
        "widths": [24, 56, 152, 368],
        "group_width": 8,
        "default_size": 224,
        "block_type": "Y",
    },
    "y004": {
        "depths": [1, 3, 6, 6],
        "widths": [48, 104, 208, 440],
        "group_width": 8,
        "default_size": 224,
        "block_type": "Y",
    },
    "y006": {
        "depths": [1, 3, 7, 4],
        "widths": [48, 112, 256, 608],
        "group_width": 16,
        "default_size": 224,
        "block_type": "Y",
    },
    "y008": {
        "depths": [1, 3, 8, 2],
        "widths": [64, 128, 320, 768],
        "group_width": 16,
        "default_size": 224,
        "block_type": "Y",
    },
    "y016": {
        "depths": [2, 6, 17, 2],
        "widths": [48, 120, 336, 888],
        "group_width": 24,
        "default_size": 224,
        "block_type": "Y",
    },
    "y032": {
        "depths": [2, 5, 13, 1],
        "widths": [72, 216, 576, 1512],
        "group_width": 24,
        "default_size": 224,
        "block_type": "Y",
    },
    "y040": {
        "depths": [2, 6, 12, 2],
        "widths": [128, 192, 512, 1088],
        "group_width": 64,
        "default_size": 224,
        "block_type": "Y",
    },
    "y064": {
        "depths": [2, 7, 14, 2],
        "widths": [144, 288, 576, 1296],
        "group_width": 72,
        "default_size": 224,
        "block_type": "Y",
    },
    "y080": {
        "depths": [2, 4, 10, 1],
        "widths": [168, 448, 896, 2016],
        "group_width": 56,
        "default_size": 224,
        "block_type": "Y",
    },
    "y120": {
        "depths": [2, 5, 11, 1],
        "widths": [224, 448, 896, 2240],
        "group_width": 112,
        "default_size": 224,
        "block_type": "Y",
    },
    "y160": {
        "depths": [2, 4, 11, 1],
        "widths": [224, 448, 1232, 3024],
        "group_width": 112,
        "default_size": 224,
        "block_type": "Y",
    },
    "y320": {
        "depths": [2, 5, 12, 1],
        "widths": [232, 696, 1392, 3712],
        "group_width": 232,
        "default_size": 224,
        "block_type": "Y",
    },
}

BASE_DOCSTRING = """This class represents the {name} architecture.

  Reference:
    - [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678)
    (CVPR 2020)

  For image classification use cases, see
  [this page for detailed examples](https://keras.io/api/applications/#usage-examples-for-image-classification-models).

  For transfer learning use cases, make sure to read the
  [guide to transfer learning & fine-tuning](https://keras.io/guides/transfer_learning/).


  The naming of models is as follows: `RegNet<block_type><flops>` where
  `block_type` is one of `(X, Y)` and `flops` signifies hundred million
  floating point operations. For example RegNetY064 corresponds to RegNet with
  Y block and 6.4 giga flops (64 hundred million flops).

  Args:
    include_rescaling: whether to rescale the inputs. If set to True,
        inputs will be passed through a `Rescaling(1/255.0)` layer.
    include_top: Whether to include the fully-connected
        layer at the top of the network.
    num_classes: Optional number of classes to classify images
        into, only to be specified if `include_top` is True.
    weights: One of `None` (random initialization), or the path to the weights
          file to be loaded, defaults to `None`.
    input_tensor: Optional Keras tensor (i.e. output of `layers.Input()`)
        to use as image input for the model.
    input_shape: Optional shape tuple, defaults to (None, None, 3).
        It should have exactly 3 inputs channels.
    pooling: Optional pooling mode for feature extraction
        when `include_top` is `False`, defaults to None.
        - `None` means that the output of the model will be
            the 4D tensor output of the
            last convolutional layer.
        - `avg` means that global average pooling
            will be applied to the output of the
            last convolutional layer, and thus
            the output of the model will be a 2D tensor.
        - `max` means that global max pooling will
            be applied.
    classifier_activation: A `str` or callable. The activation function to use
        on the "top" layer. Ignored unless `include_top=True`. Set
        `classifier_activation=None` to return the logits of the "top" layer.
        Defaults to `"softmax"`.
        When loading pretrained weights, `classifier_activation` can only
        be `None` or `"softmax"`.

  Returns:
    A `keras.Model` instance.
"""  # noqa: E501


def apply_conv2d_bn(
    x,
    filters,
    kernel_size,
    strides=1,
    use_bias=False,
    groups=1,
    padding="valid",
    kernel_initializer="he_normal",
    batch_norm=True,
    activation="relu",
    name="",
):
    x = layers.Conv2D(
        filters,
        kernel_size,
        strides=strides,
        groups=groups,
        use_bias=use_bias,
        padding=padding,
        kernel_initializer=kernel_initializer,
        name=name,
    )(x)

    if batch_norm:
        x = layers.BatchNormalization(
            momentum=0.9, epsilon=1e-5, name=name + "_bn"
        )(x)

    if activation is not None:
        x = layers.Activation(activation, name=name + f"_{activation}")(x)

    return x


def apply_stem(x, name=None):
    """Implementation of RegNet stem.

    (Common to all model variants)
    Args:
      x: Tensor, input tensor to the stem
      name: name prefix

    Returns:
      Output tensor of the Stem
    """
    if name is None:
        name = "stem" + str(backend.get_uid("stem"))

    x = apply_conv2d_bn(
        x=x,
        filters=32,
        kernel_size=(3, 3),
        strides=2,
        padding="same",
        name=name + "_stem_conv",
    )

    return x


def apply_x_block(
    inputs, filters_in, filters_out, group_width, stride=1, name=None
):
    """Implementation of X Block.
    References:
        - [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678)

    Args:
      inputs: Tensor, input tensor to the block
      filters_in: int, filters in the input tensor
      filters_out: int, filters in the output tensor
      group_width: int, group width
      stride: int (or) tuple, stride of Conv layer
      name: str, name prefix
    Returns:
      Output tensor of the block
    """
    if name is None:
        name = str(backend.get_uid("xblock"))

    if filters_in != filters_out and stride == 1:
        raise ValueError(
            f"Input filters({filters_in}) and output "
            f"filters({filters_out}) "
            f"are not equal for stride {stride}. Input and output filters "
            f"must be equal for stride={stride}."
        )

    # Declare layers
    groups = filters_out // group_width

    if stride != 1:
        skip = apply_conv2d_bn(
            x=inputs,
            filters=filters_out,
            kernel_size=(1, 1),
            strides=stride,
            activation=None,
            name=name + "_skip_1x1",
        )
    else:
        skip = inputs

    # Build block
    # conv_1x1_1
    x = apply_conv2d_bn(
        x=inputs,
        filters=filters_out,
        kernel_size=(1, 1),
        name=name + "_conv_1x1_1",
    )

    # conv_3x3
    x = apply_conv2d_bn(
        x=x,
        filters=filters_out,
        kernel_size=(3, 3),
        strides=stride,
        groups=groups,
        padding="same",
        name=name + "_conv_3x3",
    )

    # conv_1x1_2
    x = apply_conv2d_bn(
        x=x,
        filters=filters_out,
        kernel_size=(1, 1),
        activation=None,
        name=name + "_conv_1x1_2",
    )

    x = layers.Activation("relu", name=name + "_exit_relu")(x + skip)

    return x


def apply_y_block(
    inputs,
    filters_in,
    filters_out,
    group_width,
    stride=1,
    squeeze_excite_ratio=0.25,
    name=None,
):
    """Implementation of Y Block.
    References:
        - [Designing Network Design Spaces](https://arxiv.org/abs/2003.13678)

    Args:
      inputs: Tensor, input tensor to the block
      filters_in: int, filters in the input tensor
      filters_out: int, filters in the output tensor
      group_width: int, group width
      stride: int (or) tuple, stride of Conv layer
      squeeze_excite_ratio: float, expansion ratio for Squeeze and Excite block
      name: str, name prefix
    Returns:
      Output tensor of the block
    """
    if name is None:
        name = str(backend.get_uid("yblock"))

    if filters_in != filters_out and stride == 1:
        raise ValueError(
            f"Input filters({filters_in}) and output "
            f"filters({filters_out}) "
            f"are not equal for stride {stride}. Input and output filters "
            f"must be equal for stride={stride}."
        )

    groups = filters_out // group_width

    if stride != 1:
        skip = apply_conv2d_bn(
            x=inputs,
            filters=filters_out,
            kernel_size=(1, 1),
            strides=stride,
            activation=None,
            name=name + "_skip_1x1",
        )
    else:
        skip = inputs

    # Build block
    # conv_1x1_1
    x = apply_conv2d_bn(
        x=inputs,
        filters=filters_out,
        kernel_size=(1, 1),
        name=name + "_conv_1x1_1",
    )

    # conv_3x3
    x = apply_conv2d_bn(
        x=x,
        filters=filters_out,
        kernel_size=(3, 3),
        strides=stride,
        groups=groups,
        padding="same",
        name=name + "_conv_3x3",
    )

    # Squeeze-Excitation block
    x = SqueezeAndExcite2D(
        filters_out,
        bottleneck_filters=filters_out * squeeze_excite_ratio,
        name=name,
    )(x)

    # conv_1x1_2
    x = apply_conv2d_bn(
        x=x,
        filters=filters_out,
        kernel_size=(1, 1),
        activation=None,
        name=name + "_conv_1x1_2",
    )

    x = layers.Activation("relu", name=name + "_exit_relu")(x + skip)

    return x


def apply_z_block(
    inputs,
    filters_in,
    filters_out,
    group_width,
    stride=1,
    squeeze_excite_ratio=0.25,
    bottleneck_ratio=0.25,
    name=None,
):
    """Implementation of Z block.

    References:
        - [Fast and Accurate Model Scaling](https://arxiv.org/abs/2103.06877).

    Args:
      inputs: Tensor, input tensor to the block
      filters_in: int, filters in the input tensor
      filters_out: int, filters in the output tensor
      group_width: int, group width
      stride: int (or) tuple, stride
      squeeze_excite_ratio: float, expansion ration for Squeeze and Excite block
      bottleneck_ratio: float, inverted bottleneck ratio
      name: str, name prefix
    Returns:
      Output tensor of the block
    """
    if name is None:
        name = str(backend.get_uid("zblock"))

    if filters_in != filters_out and stride == 1:
        raise ValueError(
            f"Input filters({filters_in}) and output filters({filters_out})"
            f"are not equal for stride {stride}. Input and output filters "
            f"must be equal for stride={stride}."
        )

    groups = filters_out // group_width

    inv_btlneck_filters = int(filters_out / bottleneck_ratio)

    # Build block
    # conv_1x1_1
    x = apply_conv2d_bn(
        x=inputs,
        filters=inv_btlneck_filters,
        kernel_size=(1, 1),
        name=name + "_conv_1x1_1",
        activation="silu",
    )

    # conv_3x3
    x = apply_conv2d_bn(
        x=x,
        filters=inv_btlneck_filters,
        kernel_size=(3, 3),
        strides=stride,
        groups=groups,
        padding="same",
        name=name + "_conv_3x3",
        activation="silu",
    )

    # Squeeze-Excitation block
    x = SqueezeAndExcite2D(
        inv_btlneck_filters,
        bottleneck_filter=inv_btlneck_filters * squeeze_excite_ratio,
        name=name,
    )(x)

    # conv_1x1_2
    x = apply_conv2d_bn(
        x=x,
        filters=filters_out,
        kernel_size=(1, 1),
        activation=None,
        name=name + "_conv_1x1_2",
    )

    if stride != 1:
        return x
    else:
        return x + inputs


def apply_stage(
    x, block_type, depth, group_width, filters_in, filters_out, name=None
):
    """Implementation of Stage in RegNet.

    Args:
      x: Tensor, input tensor to the stage
      block_type: must be one of "X", "Y", "Z"
      depth: int, depth of stage, number of blocks to use
      group_width: int, group width of all blocks in  this stage
      filters_in: int, input filters to this stage
      filters_out: int, output filters from this stage
      name: str, name prefix

    Returns:
      Output tensor of the block
    """
    if name is None:
        name = str(backend.get_uid("stage"))

    if block_type == "X":
        x = apply_x_block(
            x,
            filters_in,
            filters_out,
            group_width,
            stride=2,
            name=f"{name}_XBlock_0",
        )
        for i in range(1, depth):
            x = apply_x_block(
                x,
                filters_out,
                filters_out,
                group_width,
                name=f"{name}_XBlock_{i}",
            )
    elif block_type == "Y":
        x = apply_y_block(
            x,
            filters_in,
            filters_out,
            group_width,
            stride=2,
            name=name + "_YBlock_0",
        )
        for i in range(1, depth):
            x = apply_y_block(
                x,
                filters_out,
                filters_out,
                group_width,
                name=f"{name}_YBlock_{i}",
            )
    elif block_type == "Z":
        x = apply_z_block(
            x,
            filters_in,
            filters_out,
            group_width,
            stride=2,
            name=f"{name}_ZBlock_0",
        )
        for i in range(1, depth):
            x = apply_z_block(
                x,
                filters_out,
                filters_out,
                group_width,
                name=f"{name}_ZBlock_{i}",
            )
    else:
        raise NotImplementedError(
            f"Block type `{block_type}` not recognized."
            f"block_type must be one of (`X`, `Y`, `Z`). "
        )
    return x


def apply_head(x, num_classes=None, name=None, activation=None):
    """Implementation of classification head of RegNet.

    Args:
      x: Tensor, input to the head block
      num_classes: int, number of classes for Dense layer
      name: str, name prefix

    Returns:
      Output logits tensor.
    """
    if name is None:
        name = str(backend.get_uid("head"))

    x = layers.GlobalAveragePooling2D(name=name + "_head_gap")(x)
    x = layers.Dense(
        num_classes, name=name + "head_dense", activation=activation
    )(x)
    return x


@keras.utils.register_keras_serializable(package="keras_cv.models")
class RegNet(keras.Model):
    """
    This class represents the architecture of RegNet
    Args:
        depths: iterable, Contains depths for each individual stages.
        widths: iterable, Contains output channel width of each individual
            stages
        group_width: int, Number of channels to be used in each group. See
            grouped convolutions for more information.
        block_type: Must be one of `{"X", "Y", "Z"}`. For more details see the
            papers "Designing network design spaces" and "Fast and Accurate
            Model Scaling"
        default_size: tuple (or) list, default input image size.
        model_name: str, An optional name for the model.
        include_rescaling: bool, whether to rescale the inputs. If set to
            True, inputs will be passed through a `Rescaling(1/255.0)` layer.
        include_top: bool, Whether to include the fully-connected
            layer at the top of the network.
        num_classes: int, Optional number of classes to classify images
            into, only to be specified if `include_top` is True, and
            if no `weights` argument is specified.
        weights: str, One of `None` (random initialization), or the path to the
            weights file to be loaded, defaults to `None`.
        input_tensor: Tensor, Optional Keras tensor (i.e. output of
            `layers.Input()`) to use as image input for the model.
        input_shape: Optional shape tuple, defaults to (None, None, 3).
            It should have exactly 3 inputs channels.
        pooling: Optional pooling mode for feature extraction
            when `include_top` is `False`, defaults to None.
            - `None` means that the output of the model will be
                the 4D tensor output of the
                last convolutional layer.
            - `avg` means that global average pooling
                will be applied to the output of the
                last convolutional layer, and thus
                the output of the model will be a 2D tensor.
            - `max` means that global max pooling will
                be applied.
        classifier_activation: A `str` or callable. The activation function to
            use on the "top" layer. Ignored unless `include_top=True`. Set
            `classifier_activation=None` to return the logits of the "top"
            layer. Defaults to `"softmax"`.
    """

    def __init__(
        self,
        depths,
        widths,
        group_width,
        block_type,
        include_rescaling,
        include_top,
        num_classes=None,
        model_name="regnet",
        weights=None,
        input_tensor=None,
        input_shape=(None, None, 3),
        pooling=None,
        classifier_activation="softmax",
        **kwargs,
    ):
        if not (weights is None or tf.io.gfile.exists(weights)):
            raise ValueError(
                "The `weights` argument should be either "
                "`None` (random initialization) "
                "or the path to the weights file to be loaded."
            )

        if include_top and not num_classes:
            raise ValueError(
                "If `include_top` is True, you should specify `num_classes`. "
                f"Received: num_classes={num_classes}"
            )

        if include_top and pooling:
            raise ValueError(
                f"`pooling` must be `None` when `include_top=True`."
                f"Received pooling={pooling} and include_top={include_top}. "
            )

        img_input = utils.parse_model_inputs(input_shape, input_tensor)
        x = img_input

        if include_rescaling:
            x = layers.Rescaling(scale=1.0 / 255.0)(x)
        x = apply_stem(x, name=model_name)

        in_channels = x.shape[-1]  # Output from Stem

        NUM_STAGES = 4

        for stage_index in range(NUM_STAGES):
            depth = depths[stage_index]
            out_channels = widths[stage_index]

            x = apply_stage(
                x,
                block_type,
                depth,
                group_width,
                in_channels,
                out_channels,
                name=model_name + "_Stage_" + str(stage_index),
            )
            in_channels = out_channels

        if include_top:
            x = apply_head(
                x, num_classes=num_classes, activation=classifier_activation
            )
        else:
            if pooling == "avg":
                x = layers.GlobalAveragePooling2D()(x)
            elif pooling == "max":
                x = layers.GlobalMaxPooling2D()(x)

        super().__init__(inputs=img_input, outputs=x, name=model_name, **kwargs)

        # Load weights.
        if weights is not None:
            self.load_weights(weights)

        self.depths = depths
        self.widths = widths
        self.group_width = group_width
        self.block_type = block_type
        self.include_rescaling = include_rescaling
        self.include_top = include_top
        self.num_classes = num_classes
        self.model_name = model_name
        self.input_tensor = input_tensor
        self.pooling = pooling
        self.classifier_activation = classifier_activation

    def get_config(self):
        return {
            "depths": self.depths,
            "widths": self.widths,
            "group_width": self.group_width,
            "block_type": self.block_type,
            "include_rescaling": self.include_rescaling,
            "include_top": self.include_top,
            "num_classes": self.num_classes,
            "model_name": self.model_name,
            "input_tensor": self.input_tensor,
            "input_shape": self.input_shape[1:],
            "pooling": self.pooling,
            "classifier_activation": self.classifier_activation,
            "trainable": self.trainable,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**config)


# Instantiating variants
def RegNetX002(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx002",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x002"]["depths"],
        MODEL_CONFIGS["x002"]["widths"],
        MODEL_CONFIGS["x002"]["group_width"],
        MODEL_CONFIGS["x002"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx002"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX004(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx004",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x004"]["depths"],
        MODEL_CONFIGS["x004"]["widths"],
        MODEL_CONFIGS["x004"]["group_width"],
        MODEL_CONFIGS["x004"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx004"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX006(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx006",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x006"]["depths"],
        MODEL_CONFIGS["x006"]["widths"],
        MODEL_CONFIGS["x006"]["group_width"],
        MODEL_CONFIGS["x006"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx006"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX008(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx008",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x008"]["depths"],
        MODEL_CONFIGS["x008"]["widths"],
        MODEL_CONFIGS["x008"]["group_width"],
        MODEL_CONFIGS["x008"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx008"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX016(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx016",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x016"]["depths"],
        MODEL_CONFIGS["x016"]["widths"],
        MODEL_CONFIGS["x016"]["group_width"],
        MODEL_CONFIGS["x016"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx016"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX032(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx032",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x032"]["depths"],
        MODEL_CONFIGS["x032"]["widths"],
        MODEL_CONFIGS["x032"]["group_width"],
        MODEL_CONFIGS["x032"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx032"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX040(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx040",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x040"]["depths"],
        MODEL_CONFIGS["x040"]["widths"],
        MODEL_CONFIGS["x040"]["group_width"],
        MODEL_CONFIGS["x040"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx040"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX064(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx064",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x064"]["depths"],
        MODEL_CONFIGS["x064"]["widths"],
        MODEL_CONFIGS["x064"]["group_width"],
        MODEL_CONFIGS["x064"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx064"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX080(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx080",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x080"]["depths"],
        MODEL_CONFIGS["x080"]["widths"],
        MODEL_CONFIGS["x080"]["group_width"],
        MODEL_CONFIGS["x080"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx080"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX120(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx120",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x120"]["depths"],
        MODEL_CONFIGS["x120"]["widths"],
        MODEL_CONFIGS["x120"]["group_width"],
        MODEL_CONFIGS["x120"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx120"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX160(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx160",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x160"]["depths"],
        MODEL_CONFIGS["x160"]["widths"],
        MODEL_CONFIGS["x160"]["group_width"],
        MODEL_CONFIGS["x160"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx160"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetX320(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnetx320",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["x320"]["depths"],
        MODEL_CONFIGS["x320"]["widths"],
        MODEL_CONFIGS["x320"]["group_width"],
        MODEL_CONFIGS["x320"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnetx320"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY002(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety002",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y002"]["depths"],
        MODEL_CONFIGS["y002"]["widths"],
        MODEL_CONFIGS["y002"]["group_width"],
        MODEL_CONFIGS["y002"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety002"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY004(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety004",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y004"]["depths"],
        MODEL_CONFIGS["y004"]["widths"],
        MODEL_CONFIGS["y004"]["group_width"],
        MODEL_CONFIGS["y004"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety004"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY006(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety006",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y006"]["depths"],
        MODEL_CONFIGS["y006"]["widths"],
        MODEL_CONFIGS["y006"]["group_width"],
        MODEL_CONFIGS["y006"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety006"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY008(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety008",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y008"]["depths"],
        MODEL_CONFIGS["y008"]["widths"],
        MODEL_CONFIGS["y008"]["group_width"],
        MODEL_CONFIGS["y008"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety008"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY016(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety016",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y016"]["depths"],
        MODEL_CONFIGS["y016"]["widths"],
        MODEL_CONFIGS["y016"]["group_width"],
        MODEL_CONFIGS["y016"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety016"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY032(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety032",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y032"]["depths"],
        MODEL_CONFIGS["y032"]["widths"],
        MODEL_CONFIGS["y032"]["group_width"],
        MODEL_CONFIGS["y032"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety032"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY040(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety040",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y040"]["depths"],
        MODEL_CONFIGS["y040"]["widths"],
        MODEL_CONFIGS["y040"]["group_width"],
        MODEL_CONFIGS["y040"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety040"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY064(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety064",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y064"]["depths"],
        MODEL_CONFIGS["y064"]["widths"],
        MODEL_CONFIGS["y064"]["group_width"],
        MODEL_CONFIGS["y064"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety064"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY080(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety080",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y080"]["depths"],
        MODEL_CONFIGS["y080"]["widths"],
        MODEL_CONFIGS["y080"]["group_width"],
        MODEL_CONFIGS["y080"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety080"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY120(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety120",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y120"]["depths"],
        MODEL_CONFIGS["y120"]["widths"],
        MODEL_CONFIGS["y120"]["group_width"],
        MODEL_CONFIGS["y120"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety120"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY160(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety160",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y160"]["depths"],
        MODEL_CONFIGS["y160"]["widths"],
        MODEL_CONFIGS["y160"]["group_width"],
        MODEL_CONFIGS["y160"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety160"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


def RegNetY320(
    *,
    include_rescaling,
    include_top,
    num_classes=None,
    weights=None,
    input_tensor=None,
    input_shape=(None, None, 3),
    pooling=None,
    model_name="regnety320",
    classifier_activation="softmax",
    **kwargs,
):
    return RegNet(
        MODEL_CONFIGS["y320"]["depths"],
        MODEL_CONFIGS["y320"]["widths"],
        MODEL_CONFIGS["y320"]["group_width"],
        MODEL_CONFIGS["y320"]["block_type"],
        model_name=model_name,
        include_top=include_top,
        include_rescaling=include_rescaling,
        weights=parse_weights(weights, include_top, "regnety320"),
        input_tensor=input_tensor,
        input_shape=input_shape,
        pooling=pooling,
        num_classes=num_classes,
        classifier_activation=classifier_activation,
        **kwargs,
    )


RegNetX002.__doc__ = BASE_DOCSTRING.format(name="RegNetX002")
RegNetX004.__doc__ = BASE_DOCSTRING.format(name="RegNetX004")
RegNetX006.__doc__ = BASE_DOCSTRING.format(name="RegNetX006")
RegNetX008.__doc__ = BASE_DOCSTRING.format(name="RegNetX008")
RegNetX016.__doc__ = BASE_DOCSTRING.format(name="RegNetX016")
RegNetX032.__doc__ = BASE_DOCSTRING.format(name="RegNetX032")
RegNetX040.__doc__ = BASE_DOCSTRING.format(name="RegNetX040")
RegNetX064.__doc__ = BASE_DOCSTRING.format(name="RegNetX064")
RegNetX080.__doc__ = BASE_DOCSTRING.format(name="RegNetX080")
RegNetX120.__doc__ = BASE_DOCSTRING.format(name="RegNetX120")
RegNetX160.__doc__ = BASE_DOCSTRING.format(name="RegNetX160")
RegNetX320.__doc__ = BASE_DOCSTRING.format(name="RegNetX320")

RegNetY002.__doc__ = BASE_DOCSTRING.format(name="RegNetY002")
RegNetY004.__doc__ = BASE_DOCSTRING.format(name="RegNetY004")
RegNetY006.__doc__ = BASE_DOCSTRING.format(name="RegNetY006")
RegNetY008.__doc__ = BASE_DOCSTRING.format(name="RegNetY008")
RegNetY016.__doc__ = BASE_DOCSTRING.format(name="RegNetY016")
RegNetY032.__doc__ = BASE_DOCSTRING.format(name="RegNetY032")
RegNetY040.__doc__ = BASE_DOCSTRING.format(name="RegNetY040")
RegNetY064.__doc__ = BASE_DOCSTRING.format(name="RegNetY064")
RegNetY080.__doc__ = BASE_DOCSTRING.format(name="RegNetY080")
RegNetY120.__doc__ = BASE_DOCSTRING.format(name="RegNetY120")
RegNetY160.__doc__ = BASE_DOCSTRING.format(name="RegNetY160")
RegNetY320.__doc__ = BASE_DOCSTRING.format(name="RegNetY320")



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy

import numpy as np

try:
    from pycocotools.coco import COCO
    from pycocotools.cocoeval import COCOeval
except ImportError:
    COCO = object
    COCOeval = None

from keras_cv.src.utils.conditional_imports import assert_pycocotools_installed

METRIC_NAMES = [
    "AP",
    "AP50",
    "AP75",
    "APs",
    "APm",
    "APl",
    "ARmax1",
    "ARmax10",
    "ARmax100",
    "ARs",
    "ARm",
    "ARl",
]


class PyCOCOWrapper(COCO):
    """COCO wrapper class.
    This class wraps COCO API object, which provides the following additional
    functionalities:
      1. Support string type image id.
      2. Support loading the groundtruth dataset using the external annotation
         dictionary.
      3. Support loading the prediction results using the external annotation
         dictionary.
    """

    def __init__(self, gt_dataset=None):
        """Instantiates a COCO-style API object.
        Args:
          eval_type: either 'box' or 'mask'.
          annotation_file: a JSON file that stores annotations of the eval
            dataset. This is required if `gt_dataset` is not provided.
          gt_dataset: the groundtruth eval dataset in COCO API format.
        """
        assert_pycocotools_installed("PyCOCOWrapper")
        COCO.__init__(self, annotation_file=None)
        self._eval_type = "box"
        if gt_dataset:
            self.dataset = gt_dataset
            self.createIndex()

    def loadRes(self, predictions):
        """Loads result file and return a result api object.
        Args:
          predictions: a list of dictionary each representing an annotation in
            COCO format. The required fields are `image_id`, `category_id`,
            `score`, `bbox`, `segmentation`.
        Returns:
          res: result COCO api object.
        Raises:
          ValueError: if the set of image id from predictions is not the subset
            of the set of image id of the groundtruth dataset.
        """
        res = COCO()
        res.dataset["images"] = copy.deepcopy(self.dataset["images"])
        res.dataset["categories"] = copy.deepcopy(self.dataset["categories"])

        image_ids = [ann["image_id"] for ann in predictions]
        if set(image_ids) != (set(image_ids) & set(self.getImgIds())):
            raise ValueError(
                "Results do not correspond to the current dataset!"
            )
        for ann in predictions:
            x1, x2, y1, y2 = [
                ann["bbox"][0],
                ann["bbox"][0] + ann["bbox"][2],
                ann["bbox"][1],
                ann["bbox"][1] + ann["bbox"][3],
            ]

            ann["area"] = ann["bbox"][2] * ann["bbox"][3]
            ann["segmentation"] = [[x1, y1, x1, y2, x2, y2, x2, y1]]

        res.dataset["annotations"] = copy.deepcopy(predictions)
        res.createIndex()
        return res


def _yxyx_to_xywh(boxes):
    if boxes.shape[-1] != 4:
        raise ValueError(
            "boxes.shape[-1] is {:d}, but must be 4.".format(boxes.shape[-1])
        )

    boxes_ymin = boxes[..., 0]
    boxes_xmin = boxes[..., 1]
    boxes_width = boxes[..., 3] - boxes[..., 1]
    boxes_height = boxes[..., 2] - boxes[..., 0]
    new_boxes = np.stack(
        [boxes_xmin, boxes_ymin, boxes_width, boxes_height], axis=-1
    )

    return new_boxes


def _convert_predictions_to_coco_annotations(predictions):
    coco_predictions = []
    num_batches = len(predictions["source_id"])
    for i in range(num_batches):
        batch_size = predictions["source_id"][i].shape[0]
        predictions["detection_boxes"][i] = predictions["detection_boxes"][
            i
        ].copy()
        for j in range(batch_size):
            max_num_detections = predictions["num_detections"][i][j]
            predictions["detection_boxes"][i][j] = _yxyx_to_xywh(
                predictions["detection_boxes"][i][j]
            )
            for k in range(max_num_detections):
                ann = {}
                ann["image_id"] = predictions["source_id"][i][j]
                ann["category_id"] = predictions["detection_classes"][i][j][k]
                ann["bbox"] = predictions["detection_boxes"][i][j][k]
                ann["score"] = predictions["detection_scores"][i][j][k]
                coco_predictions.append(ann)

    for i, ann in enumerate(coco_predictions):
        ann["id"] = i + 1

    return coco_predictions


def _convert_groundtruths_to_coco_dataset(groundtruths, label_map=None):
    source_ids = np.concatenate(groundtruths["source_id"], axis=0)
    gt_images = [{"id": i} for i in source_ids]

    gt_annotations = []
    num_batches = len(groundtruths["source_id"])
    for i in range(num_batches):
        max_num_instances = max(x.shape[0] for x in groundtruths["classes"][i])
        batch_size = groundtruths["source_id"][i].shape[0]
        for j in range(batch_size):
            num_instances = groundtruths["num_detections"][i][j]
            if num_instances > max_num_instances:
                num_instances = max_num_instances
            for k in range(int(num_instances)):
                ann = {}
                ann["image_id"] = groundtruths["source_id"][i][j]
                ann["iscrowd"] = 0
                ann["category_id"] = int(groundtruths["classes"][i][j][k])
                boxes = groundtruths["boxes"][i]
                ann["bbox"] = [
                    float(boxes[j][k][1]),
                    float(boxes[j][k][0]),
                    float(boxes[j][k][3] - boxes[j][k][1]),
                    float(boxes[j][k][2] - boxes[j][k][0]),
                ]
                ann["area"] = float(
                    (boxes[j][k][3] - boxes[j][k][1])
                    * (boxes[j][k][2] - boxes[j][k][0])
                )
                gt_annotations.append(ann)

    for i, ann in enumerate(gt_annotations):
        ann["id"] = i + 1

    if label_map:
        gt_categories = [{"id": i, "name": label_map[i]} for i in label_map]
    else:
        category_ids = [gt["category_id"] for gt in gt_annotations]
        gt_categories = [{"id": i} for i in set(category_ids)]

    gt_dataset = {
        "images": gt_images,
        "categories": gt_categories,
        "annotations": copy.deepcopy(gt_annotations),
    }
    return gt_dataset


def _concat_numpy(groundtruths, predictions):
    """Converts tensors to numpy arrays."""
    numpy_groundtruths = {}
    for key, val in groundtruths.items():
        if isinstance(val, tuple):
            val = np.concatenate(val)
        numpy_groundtruths[key] = val

    numpy_predictions = {}
    for key, val in predictions.items():
        if isinstance(val, tuple):
            val = np.concatenate(val)
        numpy_predictions[key] = val

    return numpy_groundtruths, numpy_predictions


def compute_pycoco_metrics(groundtruths, predictions):
    assert_pycocotools_installed("compute_pycoco_metrics")

    groundtruths, predictions = _concat_numpy(groundtruths, predictions)

    gt_dataset = _convert_groundtruths_to_coco_dataset(groundtruths)
    coco_gt = PyCOCOWrapper(gt_dataset=gt_dataset)
    coco_predictions = _convert_predictions_to_coco_annotations(predictions)
    coco_dt = coco_gt.loadRes(predictions=coco_predictions)
    image_ids = [ann["image_id"] for ann in coco_predictions]

    coco_eval = COCOeval(coco_gt, coco_dt, iouType="bbox")
    coco_eval.params.imgIds = image_ids
    coco_eval.evaluate()
    coco_eval.accumulate()
    coco_eval.summarize()
    coco_metrics = coco_eval.stats

    metrics = coco_metrics

    metrics_dict = {}
    for i, name in enumerate(METRIC_NAMES):
        metrics_dict[name] = metrics[i].astype(np.float32)

    return metrics_dict



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)

# In order to support both unbatched and batched inputs, the horizontal
# and vertical axis is reverse indexed
H_AXIS = -3
W_AXIS = -2


@keras_cv_export("keras_cv.layers.Rescaling")
class Rescaling(BaseImageAugmentationLayer):
    """A preprocessing layer which rescales input values to a new range.

    This layer rescales every value of an input (often an image) by multiplying
    by `scale` and adding `offset`.

    For instance:

    1. To rescale an input in the ``[0, 255]`` range
    to be in the `[0, 1]` range, you would pass `scale=1./255`.

    2. To rescale an input in the ``[0, 255]`` range to be in the `[-1, 1]`
    range, you would pass `scale=1./127.5, offset=-1`.

    Inputs can be of integer or floating point dtype, and by default the layer
    will output floats.

    Input shape:
      Arbitrary.

    Output shape:
      Same as input.

    Args:
      scale: Float, the scale to apply to the inputs.
      offset: Float, the offset to apply to the inputs.
    """

    def __init__(self, scale, offset=0.0, **kwargs):
        super().__init__(**kwargs)

        self.scale = scale
        self.offset = offset

    def augment_image(self, image, transformation, **kwargs):
        dtype = self.compute_dtype
        scale = tf.cast(self.scale, dtype)
        offset = tf.cast(self.offset, dtype)
        return tf.cast(image, dtype) * scale + offset

    def augment_label(self, label, transformation, **kwargs):
        return label

    def augment_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        return segmentation_mask

    def augment_bounding_boxes(
        self, bounding_boxes, transformation=None, **kwargs
    ):
        return bounding_boxes

    def get_config(self):
        config = {
            "scale": self.scale,
            "offset": self.offset,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops


@keras_cv_export("keras_cv.models.stable_diffusion.TextEncoder")
class TextEncoder(keras.Model):
    def __init__(
        self, max_length, vocab_size=49408, name=None, download_weights=True
    ):
        tokens = keras.layers.Input(
            shape=(max_length,), dtype="int32", name="tokens"
        )
        positions = keras.layers.Input(
            shape=(max_length,), dtype="int32", name="positions"
        )
        x = CLIPEmbedding(vocab_size, 768, max_length)([tokens, positions])
        for _ in range(12):
            x = CLIPEncoderLayer(768, 12, activation=quick_gelu)(x)
        embedded = keras.layers.LayerNormalization(epsilon=1e-5)(x)
        super().__init__([tokens, positions], embedded, name=name)

        if download_weights:
            text_encoder_weights_fpath = keras.utils.get_file(
                origin="https://huggingface.co/fchollet/stable-diffusion/resolve/main/kcv_encoder.h5",  # noqa: E501
                file_hash="4789e63e07c0e54d6a34a29b45ce81ece27060c499a709d556c7755b42bb0dc4",  # noqa: E501
            )
            self.load_weights(text_encoder_weights_fpath)


@keras_cv_export("keras_cv.models.stable_diffusion.TextEncoderV2")
class TextEncoderV2(keras.Model):
    def __init__(
        self, max_length, vocab_size=49408, name=None, download_weights=True
    ):
        tokens = keras.layers.Input(
            shape=(max_length,), dtype="int32", name="tokens"
        )
        positions = keras.layers.Input(
            shape=(max_length,), dtype="int32", name="positions"
        )
        x = CLIPEmbedding(vocab_size, 1024, max_length)([tokens, positions])
        for _ in range(23):
            x = CLIPEncoderLayer(1024, 16, activation=ops.gelu)(x)
        embedded = keras.layers.LayerNormalization(epsilon=1e-5)(x)
        super().__init__([tokens, positions], embedded, name=name)

        if download_weights:
            text_encoder_weights_fpath = keras.utils.get_file(
                origin="https://huggingface.co/ianstenbit/keras-sd2.1/resolve/main/text_encoder_v2_1.h5",  # noqa: E501
                file_hash="985002e68704e1c5c3549de332218e99c5b9b745db7171d5f31fcd9a6089f25b",  # noqa: E501
            )
            self.load_weights(text_encoder_weights_fpath)


def quick_gelu(x):
    return x * ops.sigmoid(x * 1.702)


class CLIPEmbedding(keras.layers.Layer):
    def __init__(
        self, input_dim=49408, output_dim=768, max_length=77, **kwargs
    ):
        super().__init__(**kwargs)
        self.token_embedding = keras.layers.Embedding(input_dim, output_dim)
        self.position_embedding = keras.layers.Embedding(max_length, output_dim)

    def call(self, inputs):
        tokens, positions = inputs
        tokens = self.token_embedding(tokens)
        positions = self.position_embedding(positions)
        return tokens + positions


class CLIPEncoderLayer(keras.layers.Layer):
    def __init__(self, embed_dim, num_heads, activation=None, **kwargs):
        super().__init__(**kwargs)
        self.layer_norm1 = keras.layers.LayerNormalization(epsilon=1e-5)
        self.clip_attn = CLIPAttention(embed_dim, num_heads, causal=True)
        self.layer_norm2 = keras.layers.LayerNormalization(epsilon=1e-5)
        self.fc1 = keras.layers.Dense(embed_dim * 4)
        self.fc2 = keras.layers.Dense(embed_dim)
        self.activation = activation

    def call(self, inputs):
        residual = inputs
        x = self.layer_norm1(inputs)
        x = self.clip_attn(x)
        x = residual + x
        residual = x
        x = self.layer_norm2(x)
        x = self.fc1(x)
        x = self.activation(x)
        x = self.fc2(x)
        return x + residual


class CLIPAttention(keras.layers.Layer):
    def __init__(self, embed_dim=768, num_heads=12, causal=True, **kwargs):
        super().__init__(**kwargs)
        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.causal = causal
        self.head_dim = self.embed_dim // self.num_heads
        self.scale = self.head_dim**-0.5
        self.q_proj = keras.layers.Dense(self.embed_dim)
        self.k_proj = keras.layers.Dense(self.embed_dim)
        self.v_proj = keras.layers.Dense(self.embed_dim)
        self.out_proj = keras.layers.Dense(self.embed_dim)

    def reshape_states(self, x, sequence_length, batch_size):
        x = ops.reshape(
            x, (batch_size, sequence_length, self.num_heads, self.head_dim)
        )
        return ops.transpose(
            x, (0, 2, 1, 3)
        )  # bs, heads, sequence_length, head_dim

    def call(self, inputs, attention_mask=None):
        if attention_mask is None and self.causal:
            length = ops.shape(inputs)[1]
            attention_mask = ops.triu(
                ops.ones((1, 1, length, length), dtype=self.compute_dtype)
                * -float("inf"),
                k=1,
            )

        _, tgt_len, embed_dim = inputs.shape
        query_states = self.q_proj(inputs) * self.scale
        key_states = self.reshape_states(self.k_proj(inputs), tgt_len, -1)
        value_states = self.reshape_states(self.v_proj(inputs), tgt_len, -1)

        proj_shape = (-1, tgt_len, self.head_dim)
        query_states = self.reshape_states(query_states, tgt_len, -1)
        query_states = ops.reshape(query_states, proj_shape)
        key_states = ops.reshape(key_states, proj_shape)

        src_len = tgt_len
        value_states = ops.reshape(value_states, proj_shape)
        attn_weights = query_states @ ops.transpose(key_states, (0, 2, 1))

        attn_weights = ops.reshape(
            attn_weights, (-1, self.num_heads, tgt_len, src_len)
        )
        attn_weights = attn_weights + attention_mask
        attn_weights = ops.reshape(attn_weights, (-1, tgt_len, src_len))

        attn_weights = ops.softmax(attn_weights, axis=-1)
        attn_output = attn_weights @ value_states

        attn_output = ops.reshape(
            attn_output, (-1, self.num_heads, tgt_len, self.head_dim)
        )
        attn_output = ops.transpose(attn_output, (0, 2, 1, 3))
        attn_output = ops.reshape(attn_output, (-1, tgt_len, embed_dim))
        return self.out_proj(attn_output)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.utils import preprocessing


@keras_cv_export("keras_cv.bounding_box.ensure_tensor")
def ensure_tensor(boxes, dtype=None):
    boxes = boxes.copy()
    for key in ["boxes", "classes", "confidence"]:
        if key in boxes:
            boxes[key] = preprocessing.ensure_tensor(
                boxes[key],
                dtype=dtype,
            )
    return boxes



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src import core
from keras_cv.src import layers as cv_layers
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.src.utils import fill_utils
from keras_cv.src.utils import preprocessing


def _center_crop(mask, width, height):
    masks_shape = tf.shape(mask)
    h_diff = masks_shape[0] - height
    w_diff = masks_shape[1] - width

    h_start = tf.cast(h_diff / 2, tf.int32)
    w_start = tf.cast(w_diff / 2, tf.int32)
    return tf.image.crop_to_bounding_box(mask, h_start, w_start, height, width)


@keras_cv_export("keras_cv.layers.GridMask")
class GridMask(BaseImageAugmentationLayer):
    """GridMask class for grid-mask augmentation.


    Input shape:
        Int or float tensor with values in the range [0, 255].
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `"channels_last"` format
    Output shape:
        3D (unbatched) or 4D (batched) tensor with shape:
        `(..., height, width, channels)`, in `"channels_last"` format

    Args:
        ratio_factor: A float, tuple of two floats, or `keras_cv.FactorSampler`.
            Ratio determines the ratio from spacings to grid masks.
            Lower values make the grid
            size smaller, and higher values make the grid mask large.
            Floats should be in the range [0, 1]. 0.5 indicates that grid and
            spacing will be of equal size. To always use the same value, pass a
            `keras_cv.src.ConstantFactorSampler()`.

            Defaults to `(0, 0.5)`.
        rotation_factor:
            The rotation_factor will be used to randomly rotate the grid_mask
            during training. Default to 0.1, which results in an output rotating
            by a random amount in the range [-10% * 2pi, 10% * 2pi].

            A float represented as fraction of 2 Pi, or a tuple of size 2
            representing lower and upper bound for rotating clockwise and
            counter-clockwise. A positive values means rotating counter
            clock-wise, while a negative value means clock-wise. When
            represented as a single float, this value is used for both the upper
            and lower bound. For instance, factor=(-0.2, 0.3) results in an
            output rotation by a random amount in the range [-20% * 2pi,
            30% * 2pi]. factor=0.2 results in an output rotating by a random
            amount in the range [-20% * 2pi, 20% * 2pi].

        fill_mode: Pixels inside the gridblock are filled according to the given
            mode (one of `{"constant", "gaussian_noise"}`), defaults to
            "constant".
            - *constant*: Pixels are filled with the same constant value.
            - *gaussian_noise*: Pixels are filled with random gaussian noise.
        fill_value: an integer represents of value to be filled inside the
            gridblock when `fill_mode="constant"`. Valid integer range
            [0 to 255]
        seed: Integer. Used to create a random seed.

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    random_gridmask = keras_cv.layers.preprocessing.GridMask()
    augmented_images = random_gridmask(images)
    ```

    References:
        - [GridMask paper](https://arxiv.org/abs/2001.04086)
    """

    def __init__(
        self,
        ratio_factor=(0, 0.5),
        rotation_factor=0.15,
        fill_mode="constant",
        fill_value=0.0,
        seed=None,
        **kwargs,
    ):
        super().__init__(seed=seed, **kwargs)
        self.ratio_factor = preprocessing.parse_factor(
            ratio_factor, param_name="ratio_factor"
        )

        if isinstance(rotation_factor, core.FactorSampler):
            raise ValueError(
                "Currently `GridMask.rotation_factor` does not support the "
                "`FactorSampler` API. This will be supported in the next Keras "
                "release. For now, please pass a float for the "
                "`rotation_factor` argument."
            )

        self.fill_mode = fill_mode
        self.fill_value = fill_value
        self.rotation_factor = rotation_factor
        self.random_rotate = cv_layers.RandomRotation(
            factor=rotation_factor,
            fill_mode="constant",
            fill_value=0.0,
            seed=seed,
        )
        self.auto_vectorize = False
        self._check_parameter_values()
        self.seed = seed

    def _check_parameter_values(self):
        fill_mode, fill_value = self.fill_mode, self.fill_value

        if fill_value not in range(0, 256):
            raise ValueError(
                f"fill_value should be in the range [0, 255]. Got {fill_value}"
            )

        if fill_mode not in ["constant", "gaussian_noise", "random"]:
            raise ValueError(
                '`fill_mode` should be "constant", '
                f'"gaussian_noise", or "random". Got `fill_mode`={fill_mode}'
            )

    def get_random_transformation(
        self, image=None, label=None, bounding_boxes=None, **kwargs
    ):
        ratio = self.ratio_factor()

        # compute grid mask
        input_shape = tf.shape(image)
        mask = self._compute_grid_mask(input_shape, ratio=ratio)

        # convert mask to single-channel image
        mask = tf.cast(mask, tf.float32)
        mask = tf.expand_dims(mask, axis=-1)

        # randomly rotate mask
        mask = self.random_rotate(mask)

        # compute fill
        if self.fill_mode == "constant":
            fill_value = tf.fill(input_shape, self.fill_value)
            fill_value = tf.cast(fill_value, dtype=self.compute_dtype)
        else:
            # gaussian noise
            fill_value = self._random_generator.random_normal(
                shape=input_shape, dtype=self.compute_dtype
            )

        return mask, fill_value

    def _compute_grid_mask(self, input_shape, ratio):
        height = tf.cast(input_shape[0], tf.float32)
        width = tf.cast(input_shape[1], tf.float32)

        # mask side length
        input_diagonal_len = tf.sqrt(tf.square(width) + tf.square(height))
        mask_side_len = tf.math.ceil(input_diagonal_len)

        # grid unit size
        unit_size = self._random_generator.uniform(
            shape=(),
            minval=tf.math.minimum(height * 0.5, width * 0.3),
            maxval=tf.math.maximum(height * 0.5, width * 0.3) + 1,
            dtype=tf.float32,
        )
        rectangle_side_len = tf.cast((ratio) * unit_size, tf.float32)

        # sample x and y offset for grid units randomly between 0 and unit_size
        delta_x = self._random_generator.uniform(
            shape=(), minval=0.0, maxval=unit_size, dtype=tf.float32
        )
        delta_y = self._random_generator.uniform(
            shape=(), minval=0.0, maxval=unit_size, dtype=tf.float32
        )

        # grid size (number of diagonal units in grid)
        grid_size = mask_side_len // unit_size + 1
        grid_size_range = tf.range(1, grid_size + 1)

        # diagonal corner coordinates
        unit_size_range = grid_size_range * unit_size
        x1 = unit_size_range - delta_x
        x0 = x1 - rectangle_side_len
        y1 = unit_size_range - delta_y
        y0 = y1 - rectangle_side_len

        # compute grid coordinates
        x0, y0 = tf.meshgrid(x0, y0)
        x1, y1 = tf.meshgrid(x1, y1)

        # flatten mesh grid
        x0 = tf.reshape(x0, [-1])
        y0 = tf.reshape(y0, [-1])
        x1 = tf.reshape(x1, [-1])
        y1 = tf.reshape(y1, [-1])

        # convert coordinates to mask
        corners = tf.stack([x0, y0, x1, y1], axis=-1)
        mask_side_len = tf.cast(mask_side_len, tf.int32)
        rectangle_masks = fill_utils.corners_to_mask(
            corners, mask_shape=(mask_side_len, mask_side_len)
        )
        grid_mask = tf.reduce_any(rectangle_masks, axis=0)

        return grid_mask

    def augment_image(self, image, transformation=None, **kwargs):
        mask, fill_value = transformation
        input_shape = tf.shape(image)

        # center crop mask
        input_height = input_shape[0]
        input_width = input_shape[1]
        mask = _center_crop(mask, input_width, input_height)

        # convert back to boolean mask
        mask = tf.cast(mask, tf.bool)

        return tf.where(mask, fill_value, image)

    def augment_bounding_boxes(self, bounding_boxes, **kwargs):
        return bounding_boxes

    def augment_label(self, label, transformation=None, **kwargs):
        return label

    def augment_segmentation_mask(
        self, segmentation_mask, transformation, **kwargs
    ):
        return segmentation_mask

    def get_config(self):
        config = {
            "ratio_factor": self.ratio_factor,
            "rotation_factor": self.rotation_factor,
            "fill_mode": self.fill_mode,
            "fill_value": self.fill_value,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import tf_ops
from keras_cv.src.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)
from keras_cv.src.utils import get_interpolation

H_AXIS = -3
W_AXIS = -2

supported_keys = [
    "images",
    "labels",
    "targets",
    "bounding_boxes",
    "segmentation_masks",
]


@keras_cv_export("keras_cv.layers.Resizing")
class Resizing(BaseImageAugmentationLayer):
    """A preprocessing layer which resizes images.

    This layer resizes an image input to a target height and width. The input
    should be a 4D (batched) or 3D (unbatched) tensor in `"channels_last"`
    format. Input pixel values can be of any range (e.g. `[0., 1.)` or `[0,
    255]`) and of integer or floating point dtype. By default, the layer will
    output floats.

    This layer can be called on tf.RaggedTensor batches of input images of
    distinct sizes, and will resize the outputs to dense tensors of uniform
    size.

    For an overview and full list of preprocessing layers, see the preprocessing
    [guide](https://www.tensorflow.org/guide/keras/preprocessing_layers).

    Args:
        height: Integer, the height of the output shape.
        width: Integer, the width of the output shape.
        interpolation: String, the interpolation method, defaults to
            `"bilinear"`. Supports `"bilinear"`, `"nearest"`, `"bicubic"`,
            `"area"`, `"lanczos3"`, `"lanczos5"`, `"gaussian"`,
            `"mitchellcubic"`.
        crop_to_aspect_ratio: If True, resize the images without aspect ratio
            distortion. When the original aspect ratio differs from the target
            aspect ratio, the output image will be cropped to return the largest
            possible window in the image (of size `(height, width)`) that
            matches the target aspect ratio. By default,
            (`crop_to_aspect_ratio=False`), aspect ratio may not be preserved.
        pad_to_aspect_ratio: If True, resize the images without aspect ratio
            distortion. When the original aspect ratio differs from the target
            aspect ratio, the output image will be padded to return the largest
            possible resize of the image (of size `(height, width)`) that
            matches the target aspect ratio. By default,
            (`pad_to_aspect_ratio=False`), aspect ratio may not be preserved.
        bounding_box_format: The format of bounding boxes of input dataset.
            Refer to
            https://github.com/keras-team/keras-cv/blob/master/keras_cv/bounding_box/converters.py
            for more details on supported bounding box formats.
    """

    def __init__(
        self,
        height,
        width,
        interpolation="bilinear",
        crop_to_aspect_ratio=False,
        pad_to_aspect_ratio=False,
        bounding_box_format=None,
        **kwargs,
    ):
        self.height = height
        self.width = width
        self.interpolation = interpolation
        self.crop_to_aspect_ratio = crop_to_aspect_ratio
        self.pad_to_aspect_ratio = pad_to_aspect_ratio
        self._interpolation_method = get_interpolation(interpolation)
        self.bounding_box_format = bounding_box_format
        self.force_output_dense_images = True

        if pad_to_aspect_ratio and crop_to_aspect_ratio:
            raise ValueError(
                "`Resizing()` expects at most one of `crop_to_aspect_ratio` or "
                "`pad_to_aspect_ratio` to be True."
            )

        if not pad_to_aspect_ratio and bounding_box_format:
            raise ValueError(
                "Resizing() only supports bounding boxes when in "
                "`pad_to_aspect_ratio=True` mode. "
                "Please pass `pad_to_aspect_ratio=True`"
                "when processing bounding boxes with `Resizing()`"
            )
        super().__init__(**kwargs)

    def compute_image_signature(self, images):
        return tf.TensorSpec(
            shape=(self.height, self.width, images.shape[-1]),
            dtype=self.compute_dtype,
        )

    def _augment(self, inputs):
        images = inputs.get("images", None)
        bounding_boxes = inputs.get("bounding_boxes", None)
        segmentation_masks = inputs.get("segmentation_masks", None)

        if images is not None:
            images = tf.expand_dims(images, axis=0)
            inputs["images"] = images

        if bounding_boxes is not None:
            bounding_boxes = bounding_boxes.copy()
            bounding_boxes["classes"] = tf.expand_dims(
                bounding_boxes["classes"], axis=0
            )
            bounding_boxes["boxes"] = tf.expand_dims(
                bounding_boxes["boxes"], axis=0
            )
            inputs["bounding_boxes"] = bounding_boxes

        if segmentation_masks is not None:
            segmentation_masks = tf.expand_dims(segmentation_masks, axis=0)
            inputs["segmentation_masks"] = segmentation_masks

        outputs = self._batch_augment(inputs)

        if images is not None:
            images = tf.squeeze(outputs["images"], axis=0)
            inputs["images"] = images

        if bounding_boxes is not None:
            outputs["bounding_boxes"]["classes"] = tf.squeeze(
                outputs["bounding_boxes"]["classes"], axis=0
            )
            outputs["bounding_boxes"]["boxes"] = tf.squeeze(
                outputs["bounding_boxes"]["boxes"], axis=0
            )
            inputs["bounding_boxes"] = outputs["bounding_boxes"]

        if segmentation_masks is not None:
            segmentation_masks = tf.squeeze(
                outputs["segmentation_masks"], axis=0
            )
            inputs["segmentation_masks"] = segmentation_masks

        return inputs

    def _resize_with_distortion(self, inputs):
        images = inputs.get("images", None)
        segmentation_masks = inputs.get("segmentation_masks", None)

        size = [self.height, self.width]
        images = tf.image.resize(
            images, size=size, method=self._interpolation_method
        )
        images = tf.cast(images, self.compute_dtype)

        if segmentation_masks is not None:
            segmentation_masks = tf.image.resize(
                segmentation_masks, size=size, method="nearest"
            )

        inputs["images"] = images
        inputs["segmentation_masks"] = segmentation_masks

        return inputs

    def _resize_with_pad(self, inputs):
        def resize_single_with_pad_to_aspect(x):
            image = x.get("images", None)
            bounding_boxes = x.get("bounding_boxes", None)
            segmentation_masks = x.get("segmentation_masks", None)

            # images must be dense-able at this point.
            if isinstance(image, tf.RaggedTensor):
                image = image.to_tensor()

            img_size = tf.shape(image)
            img_height = tf.cast(img_size[H_AXIS], self.compute_dtype)
            img_width = tf.cast(img_size[W_AXIS], self.compute_dtype)
            if bounding_boxes is not None:
                bounding_boxes = bounding_box.to_dense(bounding_boxes)
                bounding_boxes = bounding_box.convert_format(
                    bounding_boxes,
                    image_shape=img_size,
                    source=self.bounding_box_format,
                    target="rel_xyxy",
                )

            # how much we scale height by to hit target height
            height_scale = self.height / img_height
            width_scale = self.width / img_width

            resize_scale = tf.math.minimum(height_scale, width_scale)
            target_height = img_height * resize_scale
            target_width = img_width * resize_scale

            image = tf.image.resize(
                image,
                size=(target_height, target_width),
                method=self._interpolation_method,
            )
            if bounding_boxes is not None:
                bounding_boxes = bounding_box.convert_format(
                    bounding_boxes,
                    images=image,
                    source="rel_xyxy",
                    target="xyxy",
                )
            image = tf.image.pad_to_bounding_box(
                image, 0, 0, self.height, self.width
            )
            if bounding_boxes is not None:
                bounding_boxes = bounding_box.clip_to_image(
                    bounding_boxes, images=image, bounding_box_format="xyxy"
                )
                bounding_boxes = bounding_box.convert_format(
                    bounding_boxes,
                    images=image,
                    source="xyxy",
                    target=self.bounding_box_format,
                )
            inputs["images"] = image

            if bounding_boxes is not None:
                inputs["bounding_boxes"] = bounding_box.to_ragged(
                    bounding_boxes
                )

            if segmentation_masks is not None:
                segmentation_masks = tf.image.resize(
                    segmentation_masks,
                    size=(target_height, target_width),
                    method="nearest",
                )
                segmentation_masks = tf.image.pad_to_bounding_box(
                    tf.cast(segmentation_masks, dtype="float32"),
                    0,
                    0,
                    self.height,
                    self.width,
                )
                inputs["segmentation_masks"] = segmentation_masks

            return inputs

        size_as_shape = tf.TensorShape((self.height, self.width))
        shape = size_as_shape + inputs["images"].shape[-1:]
        img_spec = tf.TensorSpec(shape, self.compute_dtype)
        fn_output_signature = {"images": img_spec}

        bounding_boxes = inputs.get("bounding_boxes", None)
        if bounding_boxes is not None:
            boxes_spec = self._compute_bounding_box_signature(bounding_boxes)
            fn_output_signature["bounding_boxes"] = boxes_spec

        segmentation_masks = inputs.get("segmentation_masks", None)
        if segmentation_masks is not None:
            seg_map_shape = (
                size_as_shape + inputs["segmentation_masks"].shape[-1:]
            )
            seg_map_spec = tf.TensorSpec(seg_map_shape, self.compute_dtype)
            fn_output_signature["segmentation_masks"] = seg_map_spec

        return tf.map_fn(
            resize_single_with_pad_to_aspect,
            inputs,
            fn_output_signature=fn_output_signature,
        )

    def _resize_with_crop(self, inputs):
        images = inputs.get("images", None)
        bounding_boxes = inputs.get("bounding_boxes", None)
        segmentation_masks = inputs.get("segmentation_masks", None)
        if bounding_boxes is not None:
            raise ValueError(
                "Resizing(crop_to_aspect_ratio=True) does not support "
                "bounding box inputs. Please use `pad_to_aspect_ratio=True` "
                "when processing bounding boxes with Resizing()."
            )
        inputs["images"] = images
        size = [self.height, self.width]

        # tf.image.resize will always output float32 and operate more
        # efficiently on float32 unless interpolation is nearest, in which case
        # output type matches input type.
        if self.interpolation == "nearest":
            input_dtype = self.compute_dtype
        else:
            input_dtype = tf.float32

        def resize_with_crop_to_aspect(x, interpolation_method):
            if isinstance(x, tf.RaggedTensor):
                x = x.to_tensor()
            return tf_ops.smart_resize(
                x,
                size=size,
                interpolation=interpolation_method,
            )

        def resize_with_crop_to_aspect_images(x):
            return resize_with_crop_to_aspect(
                x, interpolation_method=self._interpolation_method
            )

        def resize_with_crop_to_aspect_masks(x):
            return resize_with_crop_to_aspect(x, interpolation_method="nearest")

        if isinstance(images, tf.RaggedTensor):
            size_as_shape = tf.TensorShape(size)
            shape = size_as_shape + images.shape[-1:]
            spec = tf.TensorSpec(shape, input_dtype)
            images = tf.map_fn(
                resize_with_crop_to_aspect_images,
                images,
                fn_output_signature=spec,
            )
        else:
            images = resize_with_crop_to_aspect_images(images)

        inputs["images"] = images

        if segmentation_masks is not None:
            if isinstance(segmentation_masks, tf.RaggedTensor):
                size_as_shape = tf.TensorShape(size)
                shape = size_as_shape + segmentation_masks.shape[-1:]
                spec = tf.TensorSpec(shape, input_dtype)
                segmentation_masks = tf.map_fn(
                    resize_with_crop_to_aspect_masks,
                    segmentation_masks,
                    fn_output_signature=spec,
                )
            else:
                segmentation_masks = resize_with_crop_to_aspect_masks(
                    segmentation_masks
                )

            inputs["segmentation_masks"] = segmentation_masks

        return inputs

    def _check_inputs(self, inputs):
        for key in inputs:
            if key not in supported_keys:
                raise ValueError(
                    "Resizing() currently only supports keys "
                    f"[{', '.join(supported_keys)}]. "
                    f"Key `{key}` found in inputs to `Resizing()`. "
                )

    def _batch_augment(self, inputs):
        if (
            inputs.get("bounding_boxes", None) is not None
            and self.bounding_box_format is None
        ):
            raise ValueError(
                "Resizing requires `bounding_box_format` to be set when "
                "augmenting bounding boxes, but "
                "`self.bounding_box_format=None`."
            )

        if self.crop_to_aspect_ratio:
            return self._resize_with_crop(inputs)
        if self.pad_to_aspect_ratio:
            return self._resize_with_pad(inputs)
        return self._resize_with_distortion(inputs)

    def get_config(self):
        config = {
            "height": self.height,
            "width": self.width,
            "interpolation": self.interpolation,
            "crop_to_aspect_ratio": self.crop_to_aspect_ratio,
            "pad_to_aspect_ratio": self.pad_to_aspect_ratio,
            "bounding_box_format": self.bounding_box_format,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MiT model preset configurations."""

backbone_presets_no_weights = {
    "mit_b0": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 8 transformer blocks."
            ),
            "params": 3321962,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b0/2",
    },
    "mit_b1": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 8 transformer blocks."
            ),
            "params": 13156554,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b1/2",
    },
    "mit_b2": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 16 transformer blocks."
            ),
            "params": 24201418,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b2/2",
    },
    "mit_b3": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 28 transformer blocks."
            ),
            "params": 44077258,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b3/2",
    },
    "mit_b4": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 41 transformer blocks."
            ),
            "params": 60847818,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b4/2",
    },
    "mit_b5": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 52 transformer blocks."
            ),
            "params": 81448138,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b5/2",
    },
}

backbone_presets_with_weights = {
    "mit_b0_imagenet": {
        "metadata": {
            "description": (
                "MiT (MixTransformer) model with 8 transformer blocks. Pre-trained on ImageNet-1K and scores 69% top-1 accuracy on the validation set."  # noqa: E501
            ),
            "params": 3321962,
            "official_name": "MiT",
            "path": "mit",
        },
        "kaggle_handle": "kaggle://keras/mit/keras/mit_b0_imagenet/2",
    },
}

backbone_presets = {
    **backbone_presets_no_weights,
    **backbone_presets_with_weights,
}



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf
import tree

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import config
from keras_cv.src.backend import keras
from keras_cv.src.backend import ops
from keras_cv.src.backend import scope
from keras_cv.src.utils import preprocessing

H_AXIS = -3
W_AXIS = -2

IMAGES = "images"
LABELS = "labels"
TARGETS = "targets"
BOUNDING_BOXES = "bounding_boxes"
KEYPOINTS = "keypoints"
SEGMENTATION_MASKS = "segmentation_masks"

IS_DICT = "is_dict"
BATCHED = "batched"
USE_TARGETS = "use_targets"


@keras_cv_export("keras_cv.layers.VectorizedBaseImageAugmentationLayer")
class VectorizedBaseImageAugmentationLayer(keras.layers.Layer):
    """Abstract base layer for vectorized image augmentation.

    This layer contains base functionalities for preprocessing layers which
    augment image related data, e.g. image and in the future, label and bounding
    boxes. The subclasses could avoid making certain mistakes and reduce code
    duplications.

    This layer requires you to implement one method: `augment_images()`, which
    augments one single image during the training. There are a few additional
    methods that you can implement for added functionality on the layer:

    `augment_labels()`, which handles label augmentation if the layer supports
    that.

    `augment_bounding_boxes()`, which handles the bounding box augmentation, if
    the layer supports that.

    `get_random_transformations()`, which should produce a batch of random
    transformation settings. The transformation object, which must be a batched
    Tensor or a dictionary where each input is a batched Tensor, will be passed
    to `augment_images`, `augment_labels` and `augment_bounding_boxes`, to
    coordinate the randomness behavior, eg, in the RandomFlip layer, the image
    and bounding_boxes should be changed in the same way.

    The `call()` method support two formats of inputs:
    1. Single image tensor with 3D (HWC) or 4D (NHWC) format.
    2. A dict of tensors with stable keys. The supported keys are:
      `"images"`, `"labels"` and `"bounding_boxes"` at the moment. We might add
      more keys in future when we support more types of augmentation.

    The output of the `call()` will be in two formats, which will be the same
    structure as the inputs.

    The `call()` will unpack the inputs, forward to the correct function, and
    pack the output back to the same structure as the inputs.

    By default, the dense or ragged status of the output will be preserved.
    However, you can override this behavior by setting
    `self.force_output_dense_images = True`,
    `self.force_output_dense_segmentation_masks = True` in your `__init__()`
    method. When enabled, images and segmentation masks will be converted to
    dense tensor by `to_tensor()` if ragged.

    ```python
    class SubclassLayer(VectorizedBaseImageAugmentationLayer):
      def __init__(self):
        super().__init__()
        self.force_output_dense_images = True
        self.force_output_dense_segmentation_masks = True
    ```

    Note that since the randomness is also a common functionality, this layer
    also includes a keras_backend.RandomGenerator, which can be used to
    produce the random numbers. The random number generator is stored in the
    `self._random_generator` attribute.
    """

    def __init__(self, seed=None, **kwargs):
        super().__init__(**kwargs)
        if seed:
            self._random_generator = tf.random.Generator.from_seed(seed=seed)
        else:
            self._random_generator = tf.random.get_global_generator()
        self._convert_input_args = False
        self._allow_non_tensor_positional_args = True

    @property
    def force_output_dense_images(self):
        """Control whether to force outputting of dense images."""
        return getattr(self, "_force_output_dense_images", False)

    @force_output_dense_images.setter
    def force_output_dense_images(self, force_output_dense_images):
        self._force_output_dense_images = force_output_dense_images

    @property
    def force_output_dense_segmentation_masks(self):
        """Control whether to force outputting of dense segmentation masks."""
        return getattr(self, "_force_output_dense_segmentation_masks", False)

    @force_output_dense_segmentation_masks.setter
    def force_output_dense_segmentation_masks(
        self, force_output_dense_segmentation_masks
    ):
        self._force_output_dense_segmentation_masks = (
            force_output_dense_segmentation_masks
        )

    def augment_ragged_image(self, image, transformation, **kwargs):
        """Augment an image from a ragged image batch during training.

        This method accepts a single Dense image Tensor, and returns a Dense
        image. The resulting images are then stacked back into a ragged image
        batch. The behavior of this method should be identical to that of
        `augment_images()` but is to operate on a batch-wise basis.

        Args:
            image: a single image from the batch
            transformation: a single transformation sampled from
                `get_random_transformations()`.
            kwargs: all the other call arguments (i.e. bounding_boxes, labels,
                etc.).
        Returns:
            Augmented image.
        """
        raise NotImplementedError(
            "A ragged image batch was passed to layer of type "
            f"`{type(self).__name__}`. This layer does not implement "
            "`augment_ragged_image()`. If this is a `keras_cv`, open a GitHub "
            "issue requesting Ragged functionality on the layer titled: "
            f"'`{type(self).__name__}`: ragged image support'. If this is a "
            "custom layer, implement the `augment_ragged_image()` method."
        )

    def compute_ragged_image_signature(self, images):
        """Computes the output image signature for the `augment_image()`
        function.

        Must be overridden to return tensors with different shapes than the
        input images. By default, returns either a `tf.RaggedTensorSpec`
        matching the input image spec, or a `tf.TensorSpec` matching the input
        image spec.
        """
        ragged_spec = tf.RaggedTensorSpec(
            shape=images.shape[1:],
            ragged_rank=1,
            dtype=self.compute_dtype,
        )
        return ragged_spec

    def augment_images(self, images, transformations, **kwargs):
        """Augment a batch of images during training.

        Args:
          images: 4D image input tensor to the layer. Forwarded from
            `layer.call()`. This should generally have the shape [B, H, W, C].
            Forwarded from `layer.call()`.
          transformations: The transformations object produced by
            `get_random_transformations`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 4D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_labels(self, labels, transformations, **kwargs):
        """Augment a batch of  labels during training.

        Args:
          labels: 2D label to the layer. Forwarded from `layer.call()`.
          transformations: The transformations object produced by
            `get_random_transformations`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 2D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_targets(self, targets, transformations, **kwargs):
        """Augment a batch of targets during training.

        Args:
          targets: 2D label to the layer. Forwarded from `layer.call()`.
          transformations: The transformations object produced by
            `get_random_transformations`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 2D tensor, which will be forward to `layer.call()`.
        """
        return self.augment_labels(targets, transformations, **kwargs)

    def augment_bounding_boxes(self, bounding_boxes, transformations, **kwargs):
        """Augment bounding boxes for one image during training.

        Args:
          bounding_boxes: 3D bounding boxes to the layer. Forwarded from
            `call()`.
          transformations: The transformations object produced by
            `get_random_transformations`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 3D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_keypoints(self, keypoints, transformations, **kwargs):
        """Augment a batch of keypoints for one image during training.

        Args:
          keypoints: 3D keypoints input tensor to the layer. Forwarded from
            `layer.call()`. Shape should be [batch, num_keypoints, 2] in the
            specified keypoint format.
          transformations: The transformations object produced by
            `get_random_transformations`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 3D tensor, which will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        """Augment a batch of images' segmentation masks during training.

        Args:
          segmentation_masks: 4D segmentation mask input tensor to the layer.
            This should generally have the shape [B, H, W, 1], or in some cases
            [B, H, W, C] for multilabeled data. Forwarded from `layer.call()`.
          transformations: The transformations object produced by
            `get_random_transformations`. Used to coordinate the randomness
            between image, label, bounding box, keypoints, and segmentation
            mask.

        Returns:
          output 4D tensor containing the augmented segmentation mask, which
          will be forward to `layer.call()`.
        """
        raise NotImplementedError()

    def get_random_transformation_batch(
        self,
        batch_size,
        images=None,
        labels=None,
        bounding_boxes=None,
        keypoints=None,
        segmentation_masks=None,
    ):
        """Produce random transformations config for a batch of inputs.

        This is used to produce same randomness between
        image/label/bounding_box.

        Args:
          batch_size: the batch size of transformations configuration to sample.
          images: 3D image tensor from inputs.
          labels: optional 1D label tensor from inputs.
          bounding_boxes: optional 2D bounding boxes tensor from inputs.
          segmentation_masks: optional 3D segmentation mask tensor from inputs.

        Returns:
          Any type of object, which will be forwarded to `augment_images`,
          `augment_labels` and `augment_bounding_boxes` as the `transformations`
          parameter.
        """
        # Required to work with map_fn in the ragged cast.
        return tf.zeros((batch_size))

    def _unwrap_ragged_image_call(self, inputs):
        images = inputs.get(IMAGES, None)
        labels = inputs.get(LABELS, None)
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        keypoints = inputs.get(KEYPOINTS, None)
        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)
        transformation = inputs.get("transformations")
        images = images.to_tensor()
        images = self.augment_ragged_image(
            image=images,
            label=labels,
            bounding_boxes=bounding_boxes,
            keypoints=keypoints,
            segmentation_mask=segmentation_masks,
            transformation=transformation,
        )
        return tf.RaggedTensor.from_tensor(images)

    def _batch_augment(self, inputs):
        images = inputs.get(IMAGES, None)
        raw_images = images
        labels = inputs.get(LABELS, None)
        bounding_boxes = inputs.get(BOUNDING_BOXES, None)
        keypoints = inputs.get(KEYPOINTS, None)
        segmentation_masks = inputs.get(SEGMENTATION_MASKS, None)

        batch_size = tf.shape(images)[0]

        transformations = self.get_random_transformation_batch(
            batch_size,
            images=images,
            labels=labels,
            bounding_boxes=bounding_boxes,
            keypoints=keypoints,
            segmentation_masks=segmentation_masks,
        )

        if isinstance(images, tf.RaggedTensor):
            inputs_for_raggeds = {"transformations": transformations, **inputs}
            images = tf.map_fn(
                self._unwrap_ragged_image_call,
                inputs_for_raggeds,
                fn_output_signature=self.compute_ragged_image_signature(images),
            )
        else:
            images = self.augment_images(
                images,
                transformations=transformations,
                bounding_boxes=bounding_boxes,
                labels=labels,
            )
        if (
            isinstance(images, tf.RaggedTensor)
            and self.force_output_dense_images
        ):
            images = images.to_tensor()
        result = {IMAGES: images}

        if labels is not None:
            labels = self.augment_targets(
                labels,
                transformations=transformations,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[LABELS] = labels

        if bounding_boxes is not None:
            bounding_boxes = self.augment_bounding_boxes(
                bounding_boxes,
                transformations=transformations,
                labels=labels,
                images=images,
                raw_images=raw_images,
            )
            bounding_boxes = bounding_box.to_ragged(bounding_boxes)
            result[BOUNDING_BOXES] = bounding_boxes

        if keypoints is not None:
            keypoints = self.augment_keypoints(
                keypoints,
                transformations=transformations,
                labels=labels,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            result[KEYPOINTS] = keypoints

        if segmentation_masks is not None:
            segmentation_masks = self.augment_segmentation_masks(
                segmentation_masks,
                transformations=transformations,
                labels=labels,
                bounding_boxes=bounding_boxes,
                images=images,
                raw_images=raw_images,
            )
            if (
                isinstance(segmentation_masks, tf.RaggedTensor)
                and self.force_output_dense_segmentation_masks
            ):
                segmentation_masks = segmentation_masks.to_tensor()
            result[SEGMENTATION_MASKS] = segmentation_masks

        # preserve any additional inputs unmodified by this layer.
        for key in inputs.keys() - result.keys():
            result[key] = inputs[key]
        return result

    def call(self, inputs):
        # try to convert a given backend native tensor to TensorFlow tensor
        # before passing it over to TFDataScope
        is_tf_backend = config.backend() == "tensorflow"
        is_in_tf_graph = not tf.executing_eagerly()
        contains_ragged = lambda y: any(
            tree.map_structure(
                lambda x: isinstance(x, (tf.RaggedTensor, tf.SparseTensor)),
                tree.flatten(y),
            )
        )
        inputs_contain_ragged = contains_ragged(inputs)
        if not is_tf_backend and not inputs_contain_ragged:
            inputs = tree.map_structure(
                lambda x: tf.convert_to_tensor(x), inputs
            )
        with scope.TFDataScope():
            inputs = self._ensure_inputs_are_compute_dtype(inputs)
            inputs, metadata = self._format_inputs(inputs)
            images = inputs[IMAGES]
            if images.shape.rank == 3 or images.shape.rank == 4:
                outputs = self._format_output(
                    self._batch_augment(inputs), metadata
                )
            else:
                raise ValueError(
                    "Image augmentation layers are expecting inputs to be "
                    "rank 3 (HWC) or 4D (NHWC) tensors. Got shape: "
                    f"{images.shape}"
                )
        # convert the outputs to backend native tensors if none of them
        # contain RaggedTensors. Note that if the user passed in Raggeds
        # but the outputs are dense, we still don't want to convert to
        # backend native tensors. This is to avoid breaking TF data
        # pipelines that can't easily be ported to become backend
        # agnostic.
        if not is_tf_backend and not is_in_tf_graph:
            if not inputs_contain_ragged and not contains_ragged(outputs):
                outputs = tree.map_structure(
                    # some layers return None, handle that case when
                    # converting to tensors
                    lambda x: ops.convert_to_tensor(x) if x is not None else x,
                    outputs,
                )
        return outputs

    def _format_inputs(self, inputs):
        metadata = {IS_DICT: True, USE_TARGETS: False}
        if tf.is_tensor(inputs):
            # single image input tensor
            metadata[IS_DICT] = False
            inputs = {IMAGES: inputs}
        else:
            # Copy the input dict before we mutate it.
            inputs = dict(inputs)

        metadata[BATCHED] = inputs["images"].shape.rank == 4
        if inputs["images"].shape.rank == 3:
            for key in list(inputs.keys()):
                if key == BOUNDING_BOXES:
                    inputs[BOUNDING_BOXES]["boxes"] = tf.expand_dims(
                        inputs[BOUNDING_BOXES]["boxes"], axis=0
                    )
                    inputs[BOUNDING_BOXES]["classes"] = tf.expand_dims(
                        inputs[BOUNDING_BOXES]["classes"], axis=0
                    )
                else:
                    inputs[key] = tf.expand_dims(inputs[key], axis=0)

        if not isinstance(inputs, dict):
            raise ValueError(
                "Expect the inputs to be image tensor or dict. Got "
                f"inputs={inputs}"
            )

        if BOUNDING_BOXES in inputs:
            inputs[BOUNDING_BOXES] = self._format_bounding_boxes(
                inputs[BOUNDING_BOXES]
            )

        if isinstance(inputs, dict) and TARGETS in inputs:
            # TODO(scottzhu): Check if it only contains the valid keys
            inputs[LABELS] = inputs[TARGETS]
            del inputs[TARGETS]
            metadata[USE_TARGETS] = True
            return inputs, metadata

        return inputs, metadata

    def _format_output(self, output, metadata):
        if not metadata[BATCHED]:
            for key in list(output.keys()):
                if key == BOUNDING_BOXES:
                    output[BOUNDING_BOXES]["boxes"] = tf.squeeze(
                        output[BOUNDING_BOXES]["boxes"], axis=0
                    )
                    output[BOUNDING_BOXES]["classes"] = tf.squeeze(
                        output[BOUNDING_BOXES]["classes"], axis=0
                    )
                else:
                    output[key] = tf.squeeze(output[key], axis=0)

        if not metadata[IS_DICT]:
            return output[IMAGES]
        elif metadata[USE_TARGETS]:
            output[TARGETS] = output[LABELS]
            del output[LABELS]
        return output

    def _ensure_inputs_are_compute_dtype(self, inputs):
        if not isinstance(inputs, dict):
            return preprocessing.ensure_tensor(
                inputs,
                self.compute_dtype,
            )
        # Copy the input dict before we mutate it.
        inputs = dict(inputs)
        inputs[IMAGES] = preprocessing.ensure_tensor(
            inputs[IMAGES],
            self.compute_dtype,
        )
        if LABELS in inputs:
            inputs[LABELS] = preprocessing.ensure_tensor(
                inputs[LABELS],
                self.compute_dtype,
            )
        if KEYPOINTS in inputs:
            inputs[KEYPOINTS] = preprocessing.ensure_tensor(
                inputs[KEYPOINTS],
                self.compute_dtype,
            )
        if SEGMENTATION_MASKS in inputs:
            inputs[SEGMENTATION_MASKS] = preprocessing.ensure_tensor(
                inputs[SEGMENTATION_MASKS],
                self.compute_dtype,
            )
        if BOUNDING_BOXES in inputs:
            inputs[BOUNDING_BOXES]["boxes"] = preprocessing.ensure_tensor(
                inputs[BOUNDING_BOXES]["boxes"],
                self.compute_dtype,
            )
            inputs[BOUNDING_BOXES]["classes"] = preprocessing.ensure_tensor(
                inputs[BOUNDING_BOXES]["classes"],
                self.compute_dtype,
            )
        return inputs

    def _format_bounding_boxes(self, bounding_boxes):
        # We can't catch the case where this is None, sometimes RaggedTensor
        # drops this dimension.
        if "classes" not in bounding_boxes:
            raise ValueError(
                "Bounding boxes are missing class_id. If you would like to pad "
                "the bounding boxes with class_id, use: "
                "`bounding_boxes['classes'] = "
                "tf.ones_like(bounding_boxes['boxes'])`."
            )
        return bounding_boxes



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Some code in this file was inspired & adapted from `tensorflow_models`.
# Reference:
# https://github.com/tensorflow/models/blob/master/official/vision/ops/preprocess_ops.py

import tensorflow as tf

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers.preprocessing.vectorized_base_image_augmentation_layer import (  # noqa: E501
    VectorizedBaseImageAugmentationLayer,
)
from keras_cv.src.utils import preprocessing as preprocessing_utils

H_AXIS = -3
W_AXIS = -2


@keras_cv_export("keras_cv.layers.JitteredResize")
class JitteredResize(VectorizedBaseImageAugmentationLayer):
    """JitteredResize implements resize with scale distortion.

    JitteredResize takes a three-step approach to size-distortion based image
    augmentation. This technique is specifically tuned for object detection
    pipelines. The layer takes an input of images and bounding boxes, both of
    which may be ragged. It outputs a dense image tensor, ready to feed to a
    model for training. As such this layer will commonly be the final step in an
    augmentation pipeline.

    The augmentation process is as follows:

    The image is first scaled according to a randomly sampled scale factor. The
    width and height of the image are then resized according to the sampled
    scale. This is done to introduce noise into the local scale of features in
    the image. A subset of the image is then cropped randomly according to
    `crop_size`. This crop is then padded to be `target_size`. Bounding boxes
    are translated and scaled according to the random scaling and random
    cropping.

    Args:
        target_size: A tuple representing the output size of images.
        scale_factor: A tuple of two floats or a `keras_cv.FactorSampler`. For
            each augmented image a value is sampled from the provided range.
            This factor is used to scale the input image.
            To replicate the results of the MaskRCNN paper pass `(0.8, 1.25)`.
        crop_size: (Optional) the size of the image to crop from the scaled
            image, defaults to `target_size` when not provided.
        bounding_box_format: The format of bounding boxes of input boxes.
            Refer to
            https://github.com/keras-team/keras-cv/blob/master/keras_cv/bounding_box/converters.py
            for more details on supported bounding box formats.
        interpolation: String, the interpolation method, defaults to
            `"bilinear"`. Supports `"bilinear"`, `"nearest"`, `"bicubic"`,
            `"area"`, `"lanczos3"`, `"lanczos5"`, `"gaussian"`,
            `"mitchellcubic"`.
        seed: (Optional) integer to use as the random seed.

    Example:
    ```python
    train_ds = load_object_detection_dataset()
    jittered_resize = layers.JitteredResize(
        target_size=(640, 640),
        scale_factor=(0.8, 1.25),
        bounding_box_format="xywh",
    )
    train_ds = train_ds.map(
        jittered_resize, num_parallel_calls=tf.data.AUTOTUNE
    )
    # images now are (640, 640, 3)

    # an example using crop size
    train_ds = load_object_detection_dataset()
    jittered_resize = layers.JitteredResize(
        target_size=(640, 640),
        crop_size=(250, 250),
        scale_factor=(0.8, 1.25),
        bounding_box_format="xywh",
    )
    train_ds = train_ds.map(
        jittered_resize, num_parallel_calls=tf.data.AUTOTUNE
    )
    # images now are (640, 640, 3), but they were resized from a 250x250 crop.
    ```
    """  # noqa: E501

    def __init__(
        self,
        target_size,
        scale_factor,
        crop_size=None,
        bounding_box_format=None,
        interpolation="bilinear",
        seed=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not isinstance(target_size, tuple) or len(target_size) != 2:
            raise ValueError(
                "JitteredResize() expects `target_size` to be a tuple of two "
                f"integers. Received `target_size={target_size}`"
            )

        crop_size = crop_size or target_size
        self.interpolation = preprocessing_utils.get_interpolation(
            interpolation
        )
        self.scale_factor = preprocessing_utils.parse_factor(
            scale_factor,
            min_value=0.0,
            max_value=None,
            param_name="scale_factor",
            seed=seed,
        )
        self.crop_size = crop_size
        self.target_size = target_size
        self.bounding_box_format = bounding_box_format
        self.seed = seed

        self.force_output_dense_images = True

    def compute_ragged_image_signature(self, images):
        ragged_spec = tf.RaggedTensorSpec(
            shape=list(self.target_size) + [images.shape[-1]],
            ragged_rank=1,
            dtype=self.compute_dtype,
        )
        return ragged_spec

    def get_random_transformation_batch(
        self, batch_size, images=None, **kwargs
    ):
        heights, widths = self._get_image_shape(images)
        image_shapes = tf.cast(
            tf.concat((heights, widths), axis=-1), dtype=tf.float32
        )

        scaled_sizes = tf.round(
            image_shapes * self.scale_factor(shape=(batch_size, 1))
        )
        scales = tf.where(
            tf.less(
                scaled_sizes[..., 0] / image_shapes[..., 0],
                scaled_sizes[..., 1] / image_shapes[..., 1],
            ),
            scaled_sizes[..., 0] / image_shapes[..., 0],
            scaled_sizes[..., 1] / image_shapes[..., 1],
        )

        scaled_sizes = tf.round(image_shapes * scales[..., tf.newaxis])
        image_scales = scaled_sizes / image_shapes

        max_offsets = scaled_sizes - self.crop_size
        max_offsets = tf.where(
            tf.less(max_offsets, 0), tf.zeros_like(max_offsets), max_offsets
        )
        offsets = max_offsets * self._random_generator.uniform(
            shape=(batch_size, 2), minval=0, maxval=1, dtype=tf.float32
        )
        offsets = tf.cast(offsets, tf.int32)
        return {
            "image_scales": image_scales,
            "scaled_sizes": scaled_sizes,
            "offsets": offsets,
        }

    def augment_ragged_image(self, image, transformation, **kwargs):
        image = tf.expand_dims(image, axis=0)
        scaled_sizes = transformation["scaled_sizes"]
        offsets = transformation["offsets"]
        transformation = {
            "scaled_sizes": tf.expand_dims(scaled_sizes, axis=0),
            "offsets": tf.expand_dims(offsets, axis=0),
        }
        image = self.augment_images(
            images=image, transformations=transformation, **kwargs
        )
        return tf.squeeze(image, axis=0)

    def augment_images(
        self, images, transformations, resize_method="bilinear", **kwargs
    ):
        # unpackage augmentation arguments
        scaled_sizes = transformations["scaled_sizes"]
        offsets = transformations["offsets"]
        inputs_for_resize_and_crop_single_image = {
            "images": images,
            "scaled_sizes": scaled_sizes,
            "offsets": offsets,
        }
        scaled_images = tf.map_fn(
            lambda x: self.resize_and_crop_single_image(
                x, resize_method=resize_method
            ),
            inputs_for_resize_and_crop_single_image,
            fn_output_signature=tf.float32,
        )
        return tf.cast(scaled_images, self.compute_dtype)

    def augment_segmentation_masks(
        self, segmentation_masks, transformations, **kwargs
    ):
        return self.augment_images(
            segmentation_masks, transformations, resize_method="nearest"
        )

    def augment_labels(self, labels, transformations, **kwargs):
        return labels

    def augment_bounding_boxes(
        self, bounding_boxes, transformations, raw_images=None, **kwargs
    ):
        if self.bounding_box_format is None:
            raise ValueError(
                "Please provide a `bounding_box_format` when augmenting "
                "bounding boxes with `JitteredResize()`."
            )
        if isinstance(bounding_boxes["boxes"], tf.RaggedTensor):
            bounding_boxes = bounding_box.to_dense(bounding_boxes)
        result = bounding_boxes.copy()
        image_scales = tf.cast(
            transformations["image_scales"], self.compute_dtype
        )
        offsets = tf.cast(transformations["offsets"], self.compute_dtype)

        bounding_boxes = bounding_box.convert_format(
            bounding_boxes,
            images=raw_images,
            source=self.bounding_box_format,
            target="yxyx",
        )

        # Adjusts box coordinates based on image_scale and offset.
        yxyx = bounding_boxes["boxes"]
        yxyx *= tf.tile(image_scales, [1, 2])[..., tf.newaxis, :]
        yxyx -= tf.tile(offsets, [1, 2])[..., tf.newaxis, :]

        result["boxes"] = yxyx
        result = bounding_box.clip_to_image(
            result,
            image_shape=self.target_size + (3,),
            bounding_box_format="yxyx",
        )
        result = bounding_box.convert_format(
            result,
            image_shape=self.target_size + (3,),
            source="yxyx",
            target=self.bounding_box_format,
        )
        return result

    def _get_image_shape(self, images):
        if isinstance(images, tf.RaggedTensor):
            heights = tf.reshape(images.row_lengths(), (-1, 1))
            widths = tf.reshape(
                tf.reduce_max(images.row_lengths(axis=2), 1), (-1, 1)
            )
        else:
            batch_size = tf.shape(images)[0]
            heights = tf.repeat(tf.shape(images)[H_AXIS], repeats=[batch_size])
            heights = tf.reshape(heights, shape=(-1, 1))
            widths = tf.repeat(tf.shape(images)[W_AXIS], repeats=[batch_size])
            widths = tf.reshape(widths, shape=(-1, 1))
        return tf.cast(heights, dtype=tf.int32), tf.cast(widths, dtype=tf.int32)

    def resize_and_crop_single_image(self, inputs, resize_method="bilinear"):
        image = inputs.get("images", None)
        scaled_size = inputs.get("scaled_sizes", None)
        offset = inputs.get("offsets", None)

        scaled_image = tf.image.resize(
            image, tf.cast(scaled_size, tf.int32), method=resize_method
        )
        scaled_image = scaled_image[
            offset[0] : offset[0] + self.crop_size[0],
            offset[1] : offset[1] + self.crop_size[1],
            :,
        ]
        scaled_image = tf.image.pad_to_bounding_box(
            scaled_image, 0, 0, self.target_size[0], self.target_size[1]
        )
        return scaled_image

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "target_size": self.target_size,
                "scale_factor": self.scale_factor,
                "crop_size": self.crop_size,
                "bounding_box_format": self.bounding_box_format,
                "interpolation": self.interpolation,
                "seed": self.seed,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow as tf

from keras_cv.src import bounding_box
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.layers.preprocessing.base_image_augmentation_layer import (
    BaseImageAugmentationLayer,
)


@keras_cv_export("keras_cv.layers.MixUp")
class MixUp(BaseImageAugmentationLayer):
    """MixUp implements the MixUp data augmentation technique.

    Args:
        alpha: Float between 0 and 1. Inverse scale parameter for the gamma
            distribution. This controls the shape of the distribution from which
            the smoothing values are sampled. Defaults to 0.2, which is a
            recommended value when training an imagenet1k classification model.
        seed: Integer. Used to create a random seed.

    References:
        - [MixUp paper](https://arxiv.org/abs/1710.09412).
        - [MixUp for Object Detection paper](https://arxiv.org/pdf/1902.04103).

    Example:
    ```python
    (images, labels), _ = keras.datasets.cifar10.load_data()
    images, labels = images[:10], labels[:10]
    # Labels must be floating-point and one-hot encoded
    labels = tf.cast(tf.one_hot(labels, 10), tf.float32)
    mixup = keras_cv.layers.preprocessing.MixUp(10)
    augmented_images, updated_labels = mixup(
        {'images': images, 'labels': labels}
    )
    # output == {'images': updated_images, 'labels': updated_labels}
    ```
    """

    def __init__(self, alpha=0.2, seed=None, **kwargs):
        super().__init__(seed=seed, **kwargs)
        self.alpha = alpha
        self.seed = seed

    def _sample_from_beta(self, alpha, beta, shape):
        sample_alpha = tf.random.gamma(
            shape,
            alpha=alpha,
        )
        sample_beta = tf.random.gamma(
            shape,
            alpha=beta,
        )
        return sample_alpha / (sample_alpha + sample_beta)

    def _batch_augment(self, inputs):
        self._validate_inputs(inputs)
        images = inputs.get("images", None)
        labels = inputs.get("labels", None)
        bounding_boxes = inputs.get("bounding_boxes", None)
        segmentation_masks = inputs.get("segmentation_masks", None)
        images, lambda_sample, permutation_order = self._mixup(images)
        if labels is not None:
            labels = self._update_labels(
                tf.cast(labels, dtype=self.compute_dtype),
                lambda_sample,
                permutation_order,
            )
            inputs["labels"] = labels
        if bounding_boxes is not None:
            bounding_boxes = self._update_bounding_boxes(
                bounding_boxes, permutation_order
            )
            inputs["bounding_boxes"] = bounding_boxes
        inputs["images"] = images
        if segmentation_masks is not None:
            segmentation_masks = self._update_segmentation_masks(
                segmentation_masks, lambda_sample, permutation_order
            )
            inputs["segmentation_masks"] = segmentation_masks
        return inputs

    def _augment(self, inputs):
        raise ValueError(
            "MixUp received a single image to `call`. The layer relies on "
            "combining multiple examples, and as such will not behave as "
            "expected. Please call the layer with 2 or more samples."
        )

    def _mixup(self, images):
        batch_size = tf.shape(images)[0]
        permutation_order = tf.random.shuffle(
            tf.range(0, batch_size), seed=self.seed
        )

        lambda_sample = self._sample_from_beta(
            self.alpha, self.alpha, (batch_size,)
        )
        lambda_sample = tf.cast(
            tf.reshape(lambda_sample, [-1, 1, 1, 1]), dtype=self.compute_dtype
        )

        mixup_images = tf.cast(
            tf.gather(images, permutation_order), dtype=self.compute_dtype
        )

        images = lambda_sample * images + (1.0 - lambda_sample) * mixup_images

        return images, tf.squeeze(lambda_sample), permutation_order

    def _update_labels(self, labels, lambda_sample, permutation_order):
        labels_for_mixup = tf.gather(labels, permutation_order)

        lambda_sample = tf.reshape(lambda_sample, [-1, 1])

        labels = (
            lambda_sample * labels + (1.0 - lambda_sample) * labels_for_mixup
        )

        return labels

    def _update_bounding_boxes(self, bounding_boxes, permutation_order):
        boxes, classes = bounding_boxes["boxes"], bounding_boxes["classes"]
        boxes_for_mixup = tf.gather(boxes, permutation_order)
        classes_for_mixup = tf.gather(classes, permutation_order)
        boxes = tf.concat([boxes, boxes_for_mixup], axis=1)
        classes = tf.concat([classes, classes_for_mixup], axis=1)
        return {"boxes": boxes, "classes": classes}

    def _update_segmentation_masks(
        self, segmentation_masks, lambda_sample, permutation_order
    ):
        lambda_sample = tf.reshape(lambda_sample, [-1, 1, 1, 1])

        segmentation_masks_for_mixup = tf.gather(
            segmentation_masks, permutation_order
        )

        segmentation_masks = (
            lambda_sample * segmentation_masks
            + (1.0 - lambda_sample) * segmentation_masks_for_mixup
        )

        return segmentation_masks

    def _validate_inputs(self, inputs):
        images = inputs.get("images", None)
        labels = inputs.get("labels", None)
        bounding_boxes = inputs.get("bounding_boxes", None)
        segmentation_masks = inputs.get("segmentation_masks", None)

        if images is None or (
            labels is None
            and bounding_boxes is None
            and segmentation_masks is None
        ):
            raise ValueError(
                "MixUp expects inputs in a dictionary with format "
                '{"images": images, "labels": labels}. or'
                '{"images": images, "bounding_boxes": bounding_boxes}. or'
                '{"images": images, "segmentation_masks": segmentation_masks}. '
                f"Got: inputs = {inputs}."
            )

        if labels is not None and not labels.dtype.is_floating:
            raise ValueError(
                f"MixUp received labels with type {labels.dtype}. "
                "Labels must be of type float."
            )

        if bounding_boxes is not None:
            _ = bounding_box.validate_format(bounding_boxes)

        if segmentation_masks is not None:
            if len(segmentation_masks.shape) != 4:
                raise ValueError(
                    "MixUp expects shape of segmentation_masks as "
                    "[batch, h, w, num_classes]. "
                    f"Got: shape = {segmentation_masks.shape}. "
                )

    def get_config(self):
        config = {
            "alpha": self.alpha,
            "seed": self.seed,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))



# Copyright 2023 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend import keras
from keras_cv.src.models import MiTBackbone
from keras_cv.src.models.segmentation.segformer.segformer_presets import (  # noqa: E501
    presets,
)
from keras_cv.src.models.segmentation.segformer.segformer_presets import (  # noqa: E501
    presets_with_weights,
)
from keras_cv.src.models.task import Task
from keras_cv.src.utils.python_utils import classproperty
from keras_cv.src.utils.train import get_feature_extractor


@keras_cv_export(
    ["keras_cv.models.SegFormer", "keras_cv.models.segmentation.SegFormer"]
)
class SegFormer(Task):
    """A Keras model implementing the SegFormer architecture for semantic
    segmentation.

    References:
        - [SegFormer: Simple and Efficient Design for Semantic Segmentation with Transformers](https://arxiv.org/abs/2105.15203) # noqa: E501
        - [Based on the TensorFlow implementation from DeepVision](https://github.com/DavidLandup0/deepvision/tree/main/deepvision/models/segmentation/segformer) # noqa: E501

    Args:
        backbone: `keras.Model`. The backbone network for the model that is
            used as a feature extractor for the SegFormer encoder.
            It is *intended* to be used only with the MiT backbone model which
            was created specifically for SegFormers. It should either be a
            `keras_cv.models.backbones.backbone.Backbone` or a `tf.keras.Model`
            that implements the `pyramid_level_inputs` property with keys
            "P2", "P3", "P4", and "P5" and layer names as
            values.
        num_classes: int, the number of classes for the detection model,
            including the background class.
        projection_filters: int, number of filters in the
            convolution layer projecting the concatenated features into
            a segmentation map. Defaults to 256`.

    Example:

    Using the class with a `backbone`:

    ```python
    import tensorflow as tf
    import keras_cv

    images = np.ones(shape=(1, 96, 96, 3))
    labels = np.zeros(shape=(1, 96, 96, 1))
    backbone = keras_cv.models.MiTBackbone.from_preset("mit_b0_imagenet")
    model = keras_cv.models.segmentation.SegFormer(
        num_classes=1, backbone=backbone,
    )

    # Evaluate model
    model(images)

    # Train model
    model.compile(
        optimizer="adam",
        loss=keras.losses.BinaryCrossentropy(from_logits=False),
        metrics=["accuracy"],
    )
    model.fit(images, labels, epochs=3)
    ```
    """

    def __init__(
        self,
        backbone,
        num_classes,
        projection_filters=256,
        **kwargs,
    ):
        if not isinstance(backbone, keras.layers.Layer) or not isinstance(
            backbone, keras.Model
        ):
            raise ValueError(
                "Argument `backbone` must be a `keras.layers.Layer` instance "
                f" or `keras.Model`. Received instead "
                f"backbone={backbone} (of type {type(backbone)})."
            )

        inputs = backbone.input

        feature_extractor = get_feature_extractor(
            backbone, list(backbone.pyramid_level_inputs.values())
        )
        # Multi-level dictionary
        features = list(feature_extractor(inputs).values())

        # Get H and W of level one output
        _, H, W, _ = features[0].shape
        # Project all multi-level outputs onto the same dimensionality
        # and feature map shape
        multi_layer_outs = []
        for feature_dim, feature in zip(backbone.embedding_dims, features):
            out = keras.layers.Dense(
                projection_filters, name=f"linear_{feature_dim}"
            )(feature)
            out = keras.layers.Resizing(H, W, interpolation="bilinear")(out)
            multi_layer_outs.append(out)

        # Concat now-equal feature maps
        concatenated_outs = keras.layers.Concatenate(axis=3)(
            multi_layer_outs[::-1]
        )

        # Fuse concatenated features into a segmentation map
        seg = keras.Sequential(
            [
                keras.layers.Conv2D(
                    filters=projection_filters, kernel_size=1, use_bias=False
                ),
                keras.layers.BatchNormalization(),
                keras.layers.Activation("relu"),
            ]
        )(concatenated_outs)

        seg = keras.layers.Dropout(0.1)(seg)
        seg = keras.layers.Conv2D(
            filters=num_classes, kernel_size=1, activation="softmax"
        )(seg)

        output = keras.layers.Resizing(
            height=inputs.shape[1],
            width=inputs.shape[2],
            interpolation="bilinear",
        )(seg)

        super().__init__(
            inputs=inputs,
            outputs=output,
            **kwargs,
        )

        self.num_classes = num_classes
        self.projection_filters = projection_filters
        self.backbone = backbone

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_classes": self.num_classes,
                "projection_filters": self.projection_filters,
                "backbone": keras.saving.serialize_keras_object(self.backbone),
            }
        )
        return config

    @classmethod
    def from_preset(
        cls,
        preset,
        num_classes,
        load_weights=None,
        input_shape=None,
        **kwargs,
    ):
        aliases = {
            "segformer_b0": "mit_b0",
            "segformer_b1": "mit_b1",
            "segformer_b2": "mit_b2",
            "segformer_b3": "mit_b3",
            "segformer_b4": "mit_b4",
            "segformer_b5": "mit_b5",
        }
        if preset in aliases:
            preset = aliases[preset]
        return super().from_preset(
            preset,
            load_weights=load_weights,
            num_classes=num_classes,
            input_shape=input_shape,
            **kwargs,
        )

    @classproperty
    def presets(cls):
        """Dictionary of preset names and configurations."""
        return copy.deepcopy(presets)

    @classproperty
    def presets_with_weights(cls):
        """Dictionary of preset names and configurations that include
        weights."""
        return copy.deepcopy(presets_with_weights)

    @classproperty
    def backbone_presets(cls):
        return copy.deepcopy(MiTBackbone.presets)



# Copyright 2022 The KerasCV Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

import keras_cv.src.bounding_box.validate_format as validate_format
from keras_cv.src.api_export import keras_cv_export
from keras_cv.src.backend.scope import tf_data


def _box_shape(batched, boxes_shape, max_boxes):
    # ensure we dont drop the final axis in RaggedTensor mode
    if max_boxes is None:
        shape = list(boxes_shape)
        shape[-1] = 4
        return shape
    if batched:
        return [None, max_boxes, 4]
    return [max_boxes, 4]


def _classes_shape(batched, classes_shape, max_boxes):
    if max_boxes is None:
        return None
    if batched:
        return [None, max_boxes] + classes_shape[2:]
    return [max_boxes] + classes_shape[2:]


@keras_cv_export("keras_cv.bounding_box.to_dense")
@tf_data
def to_dense(bounding_boxes, max_boxes=None, default_value=-1):
    """to_dense converts bounding boxes to Dense tensors

    Args:
        bounding_boxes: bounding boxes in KerasCV dictionary format.
        max_boxes: the maximum number of boxes, used to pad tensors to a given
            shape. This can be used to make object detection pipelines TPU
            compatible.
        default_value: the default value to pad bounding boxes with. defaults
            to -1.
    """
    info = validate_format.validate_format(bounding_boxes)

    # guards against errors in metrics regarding modification of inputs.
    # also guards against unexpected behavior when modifying downstream
    bounding_boxes = bounding_boxes.copy()

    # Already running in masked mode
    if not info["ragged"]:
        # even if already ragged, still copy the dictionary for API consistency
        return bounding_boxes

    if isinstance(bounding_boxes["classes"], tf.RaggedTensor):
        bounding_boxes["classes"] = bounding_boxes["classes"].to_tensor(
            default_value=default_value,
            shape=_classes_shape(
                info["is_batched"], bounding_boxes["classes"].shape, max_boxes
            ),
        )

    if isinstance(bounding_boxes["boxes"], tf.RaggedTensor):
        bounding_boxes["boxes"] = bounding_boxes["boxes"].to_tensor(
            default_value=default_value,
            shape=_box_shape(
                info["is_batched"], bounding_boxes["boxes"].shape, max_boxes
            ),
        )

    if "confidence" in bounding_boxes:
        if isinstance(bounding_boxes["confidence"], tf.RaggedTensor):
            bounding_boxes["confidence"] = bounding_boxes[
                "confidence"
            ].to_tensor(
                default_value=default_value,
                shape=_classes_shape(
                    info["is_batched"],
                    bounding_boxes["confidence"].shape,
                    max_boxes,
                ),
            )

    return bounding_boxes
